/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	adcbasic.c	Some routine to manage basic ADC

	When		Who	What
	25/05/18	ac	Creation
	26/01/22	ac	Better temperature from temperature sensor

	Triple mode not tested
	Interleave mode not supported
	Temperature, VBAT and VREFINT sensor implemented for F42x/ F43x
	Injected modes not supported

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2023 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aakernel.h"
#include	"gpiobasic.h"
#include	"rccbasic.h"
#include	"adcbasic.h"

#include	"stm32f4xx_ll_bus.h"

//----------------------------------------------------------------------
//	DMA descriptor for each ADC

const dmaDesc_t		adc1DmaDesc =
{
	2,										//	DMA number 1 or 2, 0 => unused descriptor
	DMA_TO_STREAMCHAN (0, 0),				//	DMA_TO_STREAMCHAN (stream number 0..7, channel number 0..7)
	ADC1_DMA_INT_PRIORITY,					//	interruptions priority
	LL_DMA_PRIORITY_HIGH					//	DMA priority
} ;

const dmaDesc_t		adc2DmaDesc =
{
	0, // 2,										//	DMA number 1 or 2, 0 => unused descriptor
	DMA_TO_STREAMCHAN (2, 1),				//	DMA_TO_STREAMCHAN (stream number 0..7, channel number 0..7)
	ADC2_DMA_INT_PRIORITY,					//	interruptions priority
	LL_DMA_PRIORITY_HIGH					//	DMA priority
} ;

#if defined (ADC3)

	const dmaDesc_t		adc3DmaDesc =
	{
		2,									//	DMA number 1 or 2, 0 => unused descriptor
		DMA_TO_STREAMCHAN (1, 2),			//	DMA_TO_STREAMCHAN (stream number 0..7, channel number 0..7)
		ADC3_DMA_INT_PRIORITY,				//	interruptions priority
		LL_DMA_PRIORITY_HIGH				//	DMA priority
	} ;

#endif	// ADC3

//----------------------------------------------------------------------
//	One descriptor for each ADC

struct adcDesc_t
{
	ADC_TypeDef		* pAdc ;	// ADC registers address
	dmaDesc_t		dmaDesc ;	// DMA descriptor
	uint32_t		flags ;		// Configuration flags
	uint32_t		triggerSource ;
	adcCallback_t	cb ;		// interruptions handler callback
	uintptr_t		cbArg ;		// interruptions handler callback user argument

} ;

adcDesc_t		adc1Desc ;
adcDesc_t		adc2Desc ;

#if defined (ADC3)
	adcDesc_t	adc3Desc ;
#endif	// ADC3

//----------------------------------------------------------------------
//	GPIO channel definition
//	No alternate pins for analog input

typedef struct
{
	uint8_t		port ;	// 'A', 'B', etc. or AA_GPIO_PORTA, etc.
	uint8_t		pin ;	// 0 to 15

} adcPin_t ;

const adcPin_t	adc12PinList [16] =		// ADC1 and ADC2 have the same pins
{
	{	'A',	0	},	// Chan 0
	{	'A',	1	},	// Chan 1
	{	'A',	2	},	// Chan 2
	{	'A',	3	},	// Chan 3
	{	'A',	4	},	// Chan 4
	{	'A',	5	},	// Chan 5
	{	'A',	6	},	// Chan 6
	{	'A',	7	},	// Chan 7
	{	'B',	0	},	// Chan 8
	{	'B',	1	},	// Chan 9
	{	'C',	0	},	// Chan 10
	{	'C',	1	},	// Chan 11
	{	'C',	2	},	// Chan 12
	{	'C',	3	},	// Chan 13
	{	'C',	4	},	// Chan 14
	{	'C',	5	},	// Chan 15
} ;

#if defined (ADC3)
const adcPin_t	adc3PinList [16] =
{
	{	'A',	0	},	// Chan 0
	{	'A',	1	},	// Chan 1
	{	'A',	2	},	// Chan 2
	{	'A',	3	},	// Chan 3
	{	'F',	6	},	// Chan 4
	{	'F',	7	},	// Chan 5
	{	'F',	8	},	// Chan 6
	{	'F',	9	},	// Chan 7
	{	'F',	10	},	// Chan 8
	{	'F',	3	},	// Chan 9
	{	'C',	0	},	// Chan 10
	{	'C',	1	},	// Chan 11
	{	'C',	2	},	// Chan 12
	{	'C',	3	},	// Chan 13
	{	'F',	4	},	// Chan 14
	{	'F',	5	},	// Chan 15
} ;
#endif	// ADC3

static	const uint32_t	regRank [16] =
{
	LL_ADC_REG_RANK_1,
	LL_ADC_REG_RANK_2,
	LL_ADC_REG_RANK_3,
	LL_ADC_REG_RANK_4,
	LL_ADC_REG_RANK_5,
	LL_ADC_REG_RANK_6,
	LL_ADC_REG_RANK_7,
	LL_ADC_REG_RANK_8,
	LL_ADC_REG_RANK_9,
	LL_ADC_REG_RANK_10,
	LL_ADC_REG_RANK_11,
	LL_ADC_REG_RANK_12,
	LL_ADC_REG_RANK_13,
	LL_ADC_REG_RANK_14,
	LL_ADC_REG_RANK_15,
	LL_ADC_REG_RANK_16,
} ;

static	uint32_t	adcCommonConfig ;		// Common configuration

//----------------------------------------------------------------------
//	Default callback for ADC interruptions handler

void adcDefaultCallback (adcDesc_t * pAdcDesc, uint32_t event, uintptr_t arg) ;

void adcDefaultCallback (adcDesc_t * pAdcDesc, uint32_t event, uintptr_t arg)
{
	(void) pAdcDesc ;
	(void) event ;
	(void) arg ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
// Initialize ADCs and some ADC common parameters
// interDelay is used only for dual or triple interleaved mode: 5 to 20

adcStatus_t		adcInit		(uint32_t flags, uint32_t prescaler, uint32_t interDelay)
{
	adcCommonConfig = flags ;		// Memorize configuration

#if ! defined (ADC2)
	(void) interDelay ;
#endif


#if defined (ADC3)
	if (flags & ADC_TRIPLE)
	{
		adcCommonConfig |= ADC_ADC1 | ADC_ADC2 ;
		#if defined (ADC3)
			adcCommonConfig |= ADC_ADC3 ;
		#endif	// ADC3
	}
#endif
#if defined (ADC2)
	if (flags & ADC_DUAL)
	{
		adcCommonConfig |= ADC_ADC1 | ADC_ADC2 ;
	}
#endif	// ADC2

	// Check parameters
#if defined (ADC3)
	if ((adcCommonConfig & (ADC_ADC1 | ADC_ADC2 | ADC_ADC3)) == 0)
#elif defined (ADC2)
	if ((adcCommonConfig & (ADC_ADC1 | ADC_ADC2)) == 0)
#else
	if ((adcCommonConfig & ADC_ADC1) == 0)
#endif
	{
		adcCommonConfig = 0 ;
		AA_ASSERT (0) ;
		return ADC_EARG ;	// No ADC specified
	}

	if (prescaler > 3)
	{
		adcCommonConfig = 0 ;
		AA_ASSERT (0) ;
		return ADC_EARG ;	// Bad prescaler
	}

#if defined (ADC2)
	if ((adcCommonConfig & ADC_INTERLEAVED) != 0  && (interDelay < 5  ||  interDelay > 20))
	{
		adcCommonConfig = 0 ;
		AA_ASSERT (0) ;
		return ADC_EARG ;	// Bas interDelay
	}
#endif	// ADC2

	// Initialize and reset ADCs
	if (adcCommonConfig & ADC_ADC1)
	{
		rccEnableClock (ADC1, RCC_APB2ENR_ADC1EN) ;
		rccResetPeriph (ADC1, RCC_APB2ENR_ADC1EN) ;
	}

#if defined (ADC2)
	if (adcCommonConfig & ADC_ADC2)
	{
		rccEnableClock (ADC2, RCC_APB2ENR_ADC2EN) ;
		rccResetPeriph (ADC2, RCC_APB2ENR_ADC2EN) ;
	}
#endif	// ADC2

#if defined (ADC3)
	if (adcCommonConfig & ADC_ADC3)
	{
		rccEnableClock (ADC3, RCC_APB2ENR_ADC3EN) ;
		rccResetPeriph (ADC3, RCC_APB2ENR_ADC3EN) ;
	}
#endif	// ADC3

	// Set CCR MULTI. Default value is independent ADCs
#if defined (ADC2)
	if (flags & ADC_DUAL)
	{
		// Dual mode for ADC1 and ADC2
		__LL_ADC_COMMON_INSTANCE (ADC1)->CCR |= LL_ADC_MULTI_DUAL_REG_SIMULT ;		// Dual: Regular simultaneous mode only
	}
#endif	// ADC2

#if defined (ADC3)
	else if (flags & ADC_TRIPLE)
	{
		// Triple mode
		__LL_ADC_COMMON_INSTANCE (ADC1)->CCR |= LL_ADC_MULTI_TRIPLE_REG_SIMULT ;	// Triple: Regular simultaneous mode only
	}
#endif	// ADC3
	else
	{
		// Only independent ADC. Nothing to do
	}

	// Set common clock
	LL_ADC_SetCommonClock (__LL_ADC_COMMON_INSTANCE (ADC1), prescaler << ADC_CCR_ADCPRE_Pos) ;

	return ADC_ENONE ;
}

//----------------------------------------------------------------------
//	Enable or disable internal sensors
//	TEMP/VREFINT and VBAT are exclusive
//	If mode is ADC_SENSOR_NONE all sensors are disabled

// For: F42x F43x F446 F401 F411  F405/415 F407/417
#define TS_CAL1_ADDR 	((uint16_t*) ((uint32_t) 0x1FFF7A2Cu)) /* 30 deg */
#define TS_CAL2_ADDR 	((uint16_t*) ((uint32_t) 0x1FFF7A2Eu)) /* 110 deg */
#define	TS_CAL_DELTA	8000		// 80C * 100
#define	TS_CAL_REF		3000		// 30C * 100

adcStatus_t		adcSensorEnable		(adcDesc_t * pAdcDesc, uint32_t mode)
{
	switch (mode)
	{
		case ADC_SENSOR_TEMP:	// Temp  or REFINT
			// VBAT dominant, TEMP recessive, =>  should remove VBATE
			__LL_ADC_COMMON_INSTANCE (ADC1)->CCR |=  ADC_CCR_TSVREFE ;
			__LL_ADC_COMMON_INSTANCE (ADC1)->CCR &= ~ADC_CCR_VBATE ;
			break ;

		case ADC_SENSOR_VBAT:
			// VBAT dominant, TEMP recessive, if both are present, VBAT is enabled
			__LL_ADC_COMMON_INSTANCE (ADC1)->CCR |=  ADC_CCR_VBATE ;
			break ;

		case ADC_SENSOR_VREF:
			// Force ADC_CCR_TSVREFE
			__LL_ADC_COMMON_INSTANCE (ADC1)->CCR |=  ADC_CCR_TSVREFE ;
			break ;

		default:	// ADC_SENSOR_NONE
			__LL_ADC_COMMON_INSTANCE (ADC1)->CCR &= ~(ADC_CCR_TSVREFE | ADC_CCR_VBATE) ;
			break ;
	}

	(void) pAdcDesc ;
	return ADC_ENONE ;
}

//----------------------------------------------------------------------
//	Convert sensor temperature value to 1/10 C, so for 40C returns 400
//	Using calibration values TS_CAL1 and TS_CAL2:
//	Temp  = 80 / (TS_CAL2 - TS_CAL1) x (VSENSE - TS_CAL1) + 30
//	The ADC must make measurements in 12 bits with VREF 3.3V,
//		because the calibration values are given for 12 bits and VREF 3.3V.

int32_t			adcToTemp			(uint32_t tsValue)
{
	int32_t		temp ;

	temp = (TS_CAL_DELTA / (* TS_CAL2_ADDR - * TS_CAL1_ADDR)) * ((int32_t) tsValue - * TS_CAL1_ADDR) + TS_CAL_REF ;
	temp /= 10 ;

	return temp ;
}

//----------------------------------------------------------------------
//	Initialize channels pins for analog use
//	Flags is ADC_ADC1 or ADC_ADC2 or ADC_ADC3
//	pChannel is a pointer to an array of values 0 to 15
//	count is the count of channels in pChannels
//	WARNING: ADC3 does not have the same pins as ADC1 and ADC2

void	adcInitChannels		(uint32_t flags, const uint8_t * pChannel, uint32_t count)
{
	const adcPin_t	* pAdcPin ;
	gpioPinDesc_t	pinDesc ;
	uint32_t		ii ;

#if defined (ADC2)
	if (flags == ADC_ADC1  ||  flags == ADC_ADC2)
#else
	if (flags == ADC_ADC1)
#endif
	{
		pAdcPin = adc12PinList;
	}
#if defined (ADC3)
	else
	{
		pAdcPin = adc3PinList;
	}
#endif	// ADC3
	pinDesc.af    = AA_GPIO_AF_0 ;
	pinDesc.flags = AA_GPIO_MODE_ANALOG ;

	for (ii = 0 ; ii < count ; ii++)
	{
		pinDesc.port =  pAdcPin [pChannel[ii]].port ;
		pinDesc.pin  =  pAdcPin [pChannel[ii]].pin ;
		gpioConfigurePin (& pinDesc) ;
	}
}

//----------------------------------------------------------------------
//	Flags is one of ADC_ADC1, ADC_ADC2, ADC_ADC3
//	and: ADC_INT_EOC or ADC_INT_EOCS, ADC_INT_OVR

adcStatus_t	adcInitAdc		(uint32_t flags, adcDesc_t ** ppAdcDesc)
{
	ADC_TypeDef		* pAdc ;
	adcDesc_t 		* pAdcDesc ;
	uint32_t		reg ;

	if ((flags & ADC_ADC1) != 0)
	{
		pAdcDesc = & adc1Desc ;
		pAdcDesc->pAdc		= ADC1 ;
		pAdcDesc->dmaDesc	= adc1DmaDesc ;
	}
#if defined (ADC2)
	else if ((flags & ADC_ADC2) != 0)
	{
		pAdcDesc = & adc2Desc ;
		pAdcDesc->pAdc		= ADC2 ;
		pAdcDesc->dmaDesc	= adc2DmaDesc ;
	}
#endif	// ADC2
#if defined (ADC3)
	else if ((flags & ADC_ADC3) != 0)
	{
		pAdcDesc = & adc3Desc ;
		pAdcDesc->pAdc		= ADC3 ;
		pAdcDesc->dmaDesc	= adc3DmaDesc ;
	}
#endif	// ADC3
	else
	{
		AA_ASSERT (0);
		return ADC_EARG ;
	}
	pAdcDesc->flags	= 0 ;
	pAdcDesc->triggerSource = 0 ;
	pAdcDesc->cb 	= adcDefaultCallback ;
	pAdcDesc->cbArg	= 0 ;

	pAdc = pAdcDesc->pAdc ;
	pAdc->CR1 &= ~0x07C0FFFFu ;		// Clear all bits of CR1
	pAdc->CR2 &= ~0x7F7F0F03u ;		// Clear all bits of CR2

	reg = pAdc->CR1 & ~(ADC_CR1_RES | LL_ADC_SEQ_SCAN_ENABLE) ;
	// Default resolution is 12 bits
	if ((flags & ADC_RESO8BITS) != 0)
	{
		reg |= LL_ADC_RESOLUTION_8B ;
	}
	pAdc->CR1 = reg ;

	reg = pAdc->CR2 & ~(ADC_CR2_ALIGN) ;
	// Default alignment is right
	if ((flags & ADC_ALIGNLEFT) != 0)
	{
		reg |= LL_ADC_DATA_ALIGN_LEFT ;
		pAdc->CR2 = reg ;
	}

	// Initialize interruptions if requested
	if (flags & (ADC_INT_EOC | ADC_INT_EOCS | ADC_INT_OVR))
	{
		// Some interrupt is requested
		NVIC_SetPriority (ADC_IRQn, ADC_INT_PRIORITY) ;
		NVIC_EnableIRQ   (ADC_IRQn) ;

		// End of sequence or end of conversion interrupt requested ?
		if (flags & (ADC_INT_EOC | ADC_INT_EOCS))
		{
			LL_ADC_EnableIT_EOCS (pAdc) ;
			if (flags & ADC_INT_EOCS)
			{
				// Interrupt at end of every channel conversion
				LL_ADC_REG_SetFlagEndOfConversion (pAdc, LL_ADC_REG_FLAG_EOC_UNITARY_CONV) ;
			}
		}

		// Overrun interrupt requested ?
		if (flags & ADC_INT_OVR)
		{
			LL_ADC_EnableIT_OVR (pAdc) ;
		}
	}

	// Initialize channels sampling time
	for (reg = 0 ; reg < 19 ; reg++)
	{
		LL_ADC_SetChannelSamplingTime (	pAdc,
										__LL_ADC_DECIMAL_NB_TO_CHANNEL (reg),
										ADC_DEF_SAMPLINGTIME) ;
	}

	* ppAdcDesc = pAdcDesc ;
	return ADC_ENONE ;
}

//----------------------------------------------------------------------
//	Flags: [ADC_DISCONT + count] | ADC_ONESHOT
//	pChannels: pointer to array of channel number 0..16, this is the conversion sequence
//	if count > 1 then scan mode is enabled

adcStatus_t	adcInitSequencer	(adcDesc_t * pAdcDesc, uint32_t flags, const uint8_t * pChannels, uint32_t count)
{
	ADC_TypeDef		* pAdc = pAdcDesc->pAdc ;
	uint32_t		reg ;
	uint32_t		ii ;

	if (count < 1 || count > 16)
	{
		AA_ASSERT (0) ;
		return ADC_EARG ;
	}

	// Set scan and discontinuous parameters
	reg = pAdc->CR1 & ~(LL_ADC_SEQ_SCAN_ENABLE | LL_ADC_REG_SEQ_DISCONT_8RANKS) ;
	// Default is no scan (conversion of 1 channel)
	if (count > 1)
	{
		reg |= LL_ADC_SEQ_SCAN_ENABLE ;

		// Default is no discontinuous
		// Discontinuous valid only if SCAN is enabled
		if ((flags & ADC_DISCONT) != 0)
		{
			reg |= ADC_CR1_DISCEN ;
			reg |= ((flags & 0x0Fu) - 1) << ADC_CR1_DISCNUM_Pos ;
		}
	}
	pAdc->CR1 = reg ;

	// Set continuous/one shot parameter
	reg = pAdc->CR2 & ~(ADC_CR2_CONT) ;
	// Default is single conversion (one shot)
	if ((flags & ADC_ONESHOT) == 0)
	{
		reg |= ADC_CR2_CONT ;
	}
	pAdc->CR2 = reg ;

	// Set channels sequence
	for (ii = 0 ; ii < count ; ii++)
	{
		LL_ADC_REG_SetSequencerRanks (	pAdc,
										regRank [ii],
										__LL_ADC_DECIMAL_NB_TO_CHANNEL (pChannels [ii])) ;
	}

	// Set scan length
	LL_ADC_REG_SetSequencerLength (pAdc, (count-1) << ADC_SQR1_L_Pos);

	return ADC_ENONE ;
}

//----------------------------------------------------------------------

adcStatus_t		adcInitSamplingTime	(adcDesc_t * pAdcDesc, const adcSamplingTime_t * pSamplingTime, uint32_t count)
{
	uint32_t		ii ;

	for (ii = 0 ; ii < count ; ii++)
	{

		LL_ADC_SetChannelSamplingTime (	pAdcDesc->pAdc,
										__LL_ADC_DECIMAL_NB_TO_CHANNEL (pSamplingTime[ii].channel),
										pSamplingTime [ii].time) ;
	}

	return ADC_ENONE ;
}

//----------------------------------------------------------------------
//	Returns a pointer to the DMA descriptor of the ADC

adcStatus_t 	adcDmaGetDesc		(adcDesc_t * pAdcDesc, dmaDesc_t ** ppDmaDesc)
{
	* ppDmaDesc = & pAdcDesc->dmaDesc ;
	return ADC_ENONE ;
}

//----------------------------------------------------------------------
//	flags is one of: ADC_DMA_SINGLE, ADC_DMA_CIRCULAR, ADC_DMA_DOUBLE

adcStatus_t	adcDmaInit	(adcDesc_t * pAdcDesc, uint32_t flags, uint32_t length)
{
	ADC_TypeDef			* pAdc = pAdcDesc->pAdc;
	const dmaDesc_t		* pDmaDesc = & pAdcDesc->dmaDesc ;
	dma_t				* pDma ;
	uint32_t			streamNum ;
	uint32_t			reg ;

	pAdcDesc->flags = flags ;			// Memorize DMA configuration for this ADC

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Set DMA transfer according to the common configuration
#if defined (ADC3)
	if ((pAdcDesc == & adc1Desc)  &&  (adcCommonConfig & ADC_DUAL) != 0)
	{
		//	DUAL => DMA mode 2
		reg = ((flags & ADC_DMA_SINGLE) != 0 ) ? LL_ADC_MULTI_REG_DMA_LIMIT_2 : LL_ADC_MULTI_REG_DMA_UNLMT_2 ;
		LL_ADC_SetMultiDMATransfer (__LL_ADC_COMMON_INSTANCE (pAdc), reg) ;
		reg = LL_DMA_PDATAALIGN_WORD ;
	}
	else
#endif	// ADC2
#if defined (ADC3)
	if ((pAdcDesc == & adc1Desc)  &&  (adcCommonConfig & ADC_TRIPLE) != 0)
	{
		//	TRIPLE => DMA mode 1
		reg = ((flags & ADC_DMA_SINGLE) != 0 ) ? LL_ADC_MULTI_REG_DMA_LIMIT_1 : LL_ADC_MULTI_REG_DMA_UNLMT_1 ;
		LL_ADC_SetMultiDMATransfer (__LL_ADC_COMMON_INSTANCE (pAdc), reg) ;
		reg = LL_DMA_PDATAALIGN_HALFWORD ;
	}
	else
#endif	// ADC3
	{
		// Set DMA for independent ADC
		LL_ADC_REG_SetDMATransfer (pAdc, ((flags & ADC_DMA_SINGLE) != 0 ) ? LL_ADC_REG_DMA_TRANSFER_LIMITED : LL_ADC_REG_DMA_TRANSFER_UNLIMITED) ;
		reg = LL_DMA_PDATAALIGN_HALFWORD ;
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Set DMA configuration
	pDma		= DMA_GETDMA (pDmaDesc->dmaNum) ;
	streamNum	= DMA_GETSTREAM (pDmaDesc) ;

	// Enable DMA clock
	rccEnableClock (pDma, 1 << ((pDma == (dma_t *) DMA1) ? RCC_AHB1ENR_DMA1EN_Pos : RCC_AHB1ENR_DMA2EN_Pos)) ;

	// Reset stream
	dmaStreamAddressRaw (pDma, streamNum)->CR = 0 ;

	LL_DMA_SetChannelSelection ((DMA_TypeDef *) pDma, streamNum, DMA_GETCHANNEL (pDmaDesc) << 25);

	LL_DMA_ConfigTransfer ((DMA_TypeDef *) pDma, streamNum,
							reg								  |		// LL_DMA_PDATAALIGN_xxx
							pDmaDesc->config                  |
							(((flags & ADC_DMA_CIRCULAR) != 0) ? LL_DMA_MODE_CIRCULAR : LL_DMA_MODE_NORMAL) |
							LL_DMA_DIRECTION_PERIPH_TO_MEMORY |
							LL_DMA_PERIPH_NOINCREMENT         |
							LL_DMA_MEMORY_INCREMENT           |
							LL_DMA_MDATAALIGN_HALFWORD) ;			// direct mode so: MSIZE is PSIZE

	// Set double buffer if required
	if ((flags & ADC_DMA_DOUBLE) != 0)
	{
		LL_DMA_EnableDoubleBufferMode ((DMA_TypeDef *) pDma, streamNum) ;
	}

	// Set DMA buffer size
	LL_DMA_SetDataLength ((DMA_TypeDef *) pDma, streamNum, length) ;

	// Configure DMA interruptions
	if ((flags & ADC_DMA_NOIT) == 0)
	{
		//	Configure NVIC for DMA transfer complete/error interruptions
		dmaSetIrq (pDmaDesc) ;

		// Enable DMA transfer interrupt and transfer error
		LL_DMA_EnableIT_TE ((DMA_TypeDef *) pDma, streamNum) ;
		LL_DMA_EnableIT_HT ((DMA_TypeDef *) pDma, streamNum) ;
		LL_DMA_EnableIT_TC ((DMA_TypeDef *) pDma, streamNum) ;
	}
	else
	{
		// Disable DMA interruptions
		LL_DMA_DisableIT_TE ((DMA_TypeDef *) pDma, streamNum) ;
		LL_DMA_DisableIT_HT ((DMA_TypeDef *) pDma, streamNum) ;
		LL_DMA_DisableIT_TC ((DMA_TypeDef *) pDma, streamNum) ;
	}
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	return ADC_ENONE ;
}

//----------------------------------------------------------------------
//	Set DMA buffers. To be called before starting ADC
//	If ADC_DMA_CIRCULAR or ADC_DMA_ONESHOT only buffer 1 is used
//	If ADC_DMA_DOUBLE and called before starting AD: buffer 1 and buffer 2 are used
//	If ADC_DMA_DOUBLE and called after starting AD: only buffer 1 is used

adcStatus_t 	adcDmaBuffer	(adcDesc_t * pAdcDesc, const void * buffer1, const void * buffer2)
{
	ADC_TypeDef			* pAdc = pAdcDesc->pAdc ;
	const dmaDesc_t 	* pDmaDesc = & pAdcDesc->dmaDesc ;
	dma_t				* pDma ;
	uint32_t			streamNum ;
	volatile uint32_t	* pDataReg ;

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
#if defined (ADC2)
	if ((pAdcDesc->pAdc == ADC1)  &&  (adcCommonConfig & ADC_DUAL) != 0)
	{
		// If DUAL use common data register
		pDataReg = & __LL_ADC_COMMON_INSTANCE(pAdc)->CDR ;
	}
	else
#endif	// ADC2
#if defined (ADC3)
	if ((pAdcDesc->pAdc == ADC1)  &&  (adcCommonConfig & ADC_TRIPLE) != 0)
	{
		// If TRIPLE use common data register
		pDataReg = & __LL_ADC_COMMON_INSTANCE(pAdc)->CDR ;
	}
	else
#endif	// ADC3
	{
		// Use ADC data register
		pDataReg = & pAdc->DR ;
	}

	if (pDmaDesc->dmaNum == DMA_NONE)
	{
		AA_ASSERT (0) ;
		return ADC_ESTATE ;
	}
	pDma		= DMA_GETDMA (pDmaDesc->dmaNum) ;
	streamNum	= DMA_GETSTREAM (pDmaDesc) ;

	// Set DMA addresses
	if ((LL_ADC_IsEnabled (pAdc)) == 0)
	{
		// ADC not enabled: initialize DMA buffers
		LL_DMA_SetMemoryAddress ((DMA_TypeDef *) pDma, streamNum, (uint32_t) buffer1) ;
		LL_DMA_SetPeriphAddress ((DMA_TypeDef *) pDma, streamNum, (uint32_t) pDataReg) ;

		if ((pAdcDesc->flags & ADC_DMA_DOUBLE) != 0)
		{
			// Also set second buffer address
			LL_DMA_SetMemory1Address ((DMA_TypeDef *) pDma, streamNum, (uint32_t) buffer2) ;
		}
	}
	else
	{
		// ADC enabled : update memory DMA buffer not in use (double buffer only)
		if (LL_DMA_GetCurrentTargetMem ((DMA_TypeDef *) pDma, streamNum) == 0)
		{
			LL_DMA_SetMemory1Address ((DMA_TypeDef *) pDma, streamNum, (uint32_t) buffer1) ;
		}
		else
		{
			LL_DMA_SetMemoryAddress ((DMA_TypeDef *) pDma, streamNum, (uint32_t) buffer1) ;
		}
	}

	return ADC_ENONE ;
}

//----------------------------------------------------------------------
//	triggerSource: one of LL_ADC_REG_TRIG_xxx
//	edge: one of LL_ADC_REG_TRIG_EXT_RISING, LL_ADC_REG_TRIG_EXT_FALLING, LL_ADC_REG_TRIG_EXT_RISINGFALLING
//	edge: ignored if triggerSource is LL_ADC_REG_TRIG_SOFTWARE


adcStatus_t	adcSetTrigger	(adcDesc_t * pAdcDesc, uint32_t triggerSource, uint32_t edge)
{
	uint32_t	reg = triggerSource ;

	if (triggerSource != LL_ADC_REG_TRIG_SOFTWARE)
	{
		// build ADC_CR2 EXTEN:EXTSEL fields
		reg = (triggerSource & ~LL_ADC_REG_TRIG_EXT_RISINGFALLING) | edge ;
	}
	pAdcDesc->triggerSource = reg ;

	return ADC_ENONE ;
}

//----------------------------------------------------------------------
//	Start ADC conversion

adcStatus_t adcStart		(adcDesc_t * pAdcDesc)
{
	ADC_TypeDef			* pAdc = pAdcDesc->pAdc ;
	const dmaDesc_t 	* pDmaDesc = & pAdcDesc->dmaDesc ;
	dma_t				* pDma ;
	uint32_t			streamNum ;

	// Enable the DMA transfer
	if (pDmaDesc->dmaNum != DMA_NONE)
	{
		pDma		= DMA_GETDMA (pDmaDesc->dmaNum) ;
		streamNum	= DMA_GETSTREAM (pDmaDesc) ;
		LL_DMA_EnableStream ((DMA_TypeDef *) pDma, streamNum) ;
	}

	// Enable ADC channel
	LL_ADC_Enable (pAdc) ;

	// Enable non software trigger
	if (pAdcDesc->triggerSource != LL_ADC_REG_TRIG_SOFTWARE)
	{
		uint32_t	reg ;
		reg = pAdc->CR2 & ~(ADC_CR2_EXTEN | ADC_CR2_EXTSEL) ;
		pAdc->CR2 = reg |pAdcDesc->triggerSource ;
	}

	return ADC_ENONE ;
}

//----------------------------------------------------------------------
//	Stop ADC conversion

adcStatus_t adcStop		(adcDesc_t * pAdcDesc)
{
	ADC_TypeDef				* pAdc = pAdcDesc->pAdc ;
	const dmaDesc_t 	* pDmaDesc = & pAdcDesc->dmaDesc ;

	LL_ADC_REG_StopConversionExtTrig (pAdc) ;

	LL_ADC_Disable (pAdc) ;

	if(pDmaDesc->dmaNum != DMA_NONE)
	{
		dma_t			* pDma ;
		uint32_t		streamNum ;

		pDma		= DMA_GETDMA (pDmaDesc->dmaNum) ;
		streamNum	= DMA_GETSTREAM (pDmaDesc) ;
		LL_DMA_DisableStream ((DMA_TypeDef *) pDma, streamNum) ;
	}

	return ADC_ENONE ;
}

//----------------------------------------------------------------------
// Trigger ADC conversion by software

void 	adcSwTrigger	(adcDesc_t * pAdcDesc)
{
	LL_ADC_REG_StartConversionSWStart (pAdcDesc->pAdc) ;
}

//----------------------------------------------------------------------
//	ADC interruptions callback
//	Use dmaSetCallback() for DMA callback

adcStatus_t	adcSetCallback	(adcDesc_t * pAdcDesc, adcCallback_t cb, uintptr_t cbArg)
{
	pAdcDesc->cb = cb ;
	pAdcDesc->cbArg = cbArg ;

	return ADC_ENONE ;
}

//----------------------------------------------------------------------
//	Allows to wait for end of unitary conversion or sequence
//	Returns ADC_EFAIL if not ended
//	Returns ADC_ENONE if ended. EOCS flag is cleared

adcStatus_t	adcWaitForEoc	(adcDesc_t * pAdcDesc)
{
	if (LL_ADC_IsActiveFlag_EOCS (pAdcDesc->pAdc) == 0)
	{
		return ADC_EFAIL;
	}
	LL_ADC_ClearFlag_EOCS (pAdcDesc->pAdc) ;
	return ADC_ENONE ;
}

//----------------------------------------------------------------------
//	Allows to read unitary conversion data
//	Independent ADC only
// TODO: read DUAL or TRIPLE ?

void	adcReadUnitary	(adcDesc_t * pAdcDesc, uint16_t * pData)
{
	* pData = LL_ADC_REG_ReadConversionData12 (pAdcDesc->pAdc) ;
}

//----------------------------------------------------------------------
//	One IRQ common to all ADCs : ADC_IRQn

void	ADC_IRQHandler	(void) ;
void	ADC_IRQHandler	(void)
{
//	ADC_TypeDef		* pAdc = ADC1 ;

	aaIntEnter () ;

	// Check for ADC1 interruptions	// **********************************************
	if (LL_ADC_IsActiveFlag_EOCS (ADC1) != 0)
	{
		// Clear flag ADC group regular end of unitary conversion
		LL_ADC_ClearFlag_EOCS (ADC1) ;

		// Call interrupt treatment function
		(* adc1Desc.cb) (& adc1Desc, ADC_EVENT_EOC, adc1Desc.cbArg) ;
	}

	// Check whether ADC group regular overrun caused the ADC interrupt
	if (LL_ADC_IsActiveFlag_OVR (ADC1) != 0)
	{
		// Clear flag ADC group regular overrun
		LL_ADC_ClearFlag_OVR (ADC1) ;

		// Call interrupt treatment function
		(* adc1Desc.cb) (& adc1Desc, ADC_EVENT_OVR, adc1Desc.cbArg) ;
	}

	// Check for ADC2 interruptions	// **********************************************
#if defined (ADC2)
	if (LL_ADC_IsActiveFlag_EOCS (ADC2) != 0)
	{
		// Clear flag ADC group regular end of unitary conversion
		LL_ADC_ClearFlag_EOCS (ADC2) ;

		// Call interrupt treatment function
		(* adc2Desc.cb) (& adc2Desc, ADC_EVENT_EOC, adc2Desc.cbArg) ;
	}

	// Check whether ADC group regular overrun caused the ADC interrupt
	if (LL_ADC_IsActiveFlag_OVR (ADC2) != 0)
	{
		// Clear flag ADC group regular overrun
		LL_ADC_ClearFlag_OVR (ADC2) ;

		// Call interrupt treatment function
		(* adc2Desc.cb) (& adc2Desc, ADC_EVENT_OVR, adc2Desc.cbArg) ;
	}
#endif	// ADC2

#if defined (ADC3)
	// Check for ADC3 interruptions	// **********************************************
	if (LL_ADC_IsActiveFlag_EOCS (ADC3) != 0)
	{
		// Clear flag ADC group regular end of unitary conversion
		LL_ADC_ClearFlag_EOCS (ADC3) ;

		// Call interrupt treatment function
		(* adc3Desc.cb) (& adc3Desc, ADC_EVENT_EOC, adc3Desc.cbArg) ;
	}

	// Check whether ADC group regular overrun caused the ADC interrupt
	if (LL_ADC_IsActiveFlag_OVR (ADC3) != 0)
	{
		// Clear flag ADC group regular overrun
		LL_ADC_ClearFlag_OVR (ADC3) ;

		// Call interrupt treatment function
		(* adc3Desc.cb) (& adc3Desc, ADC_EVENT_OVR, adc3Desc.cbArg) ;
	}
#endif	// ADC3

	aaIntExit () ;
}

//----------------------------------------------------------------------
