/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	dacbasic.c	DAC management with DMA (1 DAC with 2 channels)

	When		Who	What
	05/05/18	ac	Creation
	16/03/20	ac	Add dual 8 bit mode.

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2023 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aakernel.h"
#include	"dacbasic.h"
#include	"rccbasic.h"
#include	"aalogmes.h"

#if (defined DAC  ||  defined DAC1)
//----------------------------------------------------------------------

// Include the proper DAC descriptors
#if (WITH_DAC_X == 1)

	#if (defined STM32F405xx || defined STM32F415xx || defined STM32F407xx || defined STM32F417xx ||	\
		 defined STM32F427xx || defined STM32F437xx || defined STM32F429xx || defined STM32F439xx ||	\
		 defined STM32F469xx || defined STM32F479xx || defined STM32F446xx || \
		 defined STM32F413xx || defined STM32F423xx)
		#include "dacF4xx.h"
	#endif

	// Count of DAC devices in dacDescArray
	STATIC	const uint32_t	dacArrayCount = sizeof (dacDescArray) / sizeof (dacDesc_t) ;

#endif	// WITH_DAC_X

//----------------------------------------------------------------------
//	DAC I/O management descriptor

typedef struct
{
	const dacDesc_t	* pDacDesc ;	// DAC descriptor
	uint32_t		flags ;			// Common configuration flags
	uint32_t		flagsChan [2] ;	// Channels configuration flags
	uint32_t		dataReg [2] ;	// Data register offset (DMA_SxPAR)
	dmaCallback_t	cb [2] ;		// DAC callback
	uintptr_t		cbArg [2] ; 	// DAC callback arg

} dacIoDesc_t ;

STATIC	dacIoDesc_t		ioDesc1 ;	// IO Descriptor for DAC1
STATIC	dacIoDesc_t		* pIoDesc = & ioDesc1 ;

//----------------------------------------------------------------------
//	Use id to get a DAC descriptor from dacDescArray

dacStatus_t		dacGetDesc (uint32_t id, const dacDesc_t ** ppDacDesc)
{
	#if (WITH_DAC_X == 1)
	{
		uint32_t	ii ;

		// Find DAC descriptor from id
		for (ii = 0u ; ii < dacArrayCount ; ii++)
		{
			if (dacDescArray [ii].id == id)
			{
				* ppDacDesc = & dacDescArray [ii] ;
				return DAC_ENONE ;
			}
		}
	}
	#else
	{
		(void) id ;
		(void) ppDacDesc ;
	}
	#endif

	return DAC_EFAIL ;	// No DAC supported
}
//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	Initialize DAC common parameters
//	Flags is DAC_DUAL or a combination of DAC_MONO1 and DAC_MONO2

dacStatus_t 	dacInit			(const dacDesc_t * pDacDesc, uint32_t flags)
{
	DAC_TypeDef		* pDac ;

	pDac = pDacDesc->pDac ;

	// Set DAC IO descriptor
	pIoDesc->pDacDesc	= pDacDesc ;
	pIoDesc->flags		= flags ;

	// Enable the peripheral clock for DAC
	rccEnableClock (pDac, 1u << pDacDesc->periphEnPos) ;

	// Reset peripheral => all configuration registers to reset values
	rccResetPeriph (pDac, 1u << pDacDesc->periphRstPos) ;

	//	Configure output pins
	if ((flags & (DAC_MONO1 | DAC_DUAL)) != 0u)
	{
		gpioConfigurePin (& pDacDesc->dacOut1) ;
	}
	if ((flags & (DAC_MONO2 | DAC_DUAL)) != 0u)
	{
		gpioConfigurePin (& pDacDesc->dacOut2) ;
	}

	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	Initialize one channel parameters.
//	flags is a combination of:
//		DAC_OUTBUFFDIS / DAC_OUTBUFFEN
//		DAC_BITS8      / DAC_BITS12
//		DAC_ALIGNLEFT  / DAC_ALIGNRIGHT

//	trigSrc is one of LL_DAC_TRIG_xxx:
//		LL_DAC_TRIG_SOFTWARE
//		LL_DAC_TRIG_EXT_TIM2_TRGO
//		LL_DAC_TRIG_EXT_TIM8_TRGO
//		LL_DAC_TRIG_EXT_TIM4_TRGO
//		LL_DAC_TRIG_EXT_TIM6_TRGO
//		LL_DAC_TRIG_EXT_TIM7_TRGO
//		LL_DAC_TRIG_EXT_TIM5_TRGO
//		LL_DAC_TRIG_EXT_EXTI_LINE9

dacStatus_t 	dacSetChannel		(uint32_t channel, uint32_t flags, uint32_t trigSrc)
{
	uint32_t		llChannel ;
	DAC_TypeDef		* pDac = pIoDesc->pDacDesc->pDac ;
	uint32_t		* pFlagsChan ;
	uint32_t		* pDataReg ;

	// Memorize configuration
	if (channel == DAC_CHAN1)
	{
		llChannel = LL_DAC_CHANNEL_1 ;
		pFlagsChan = & pIoDesc->flagsChan [0] ;
	}
	else if (channel == DAC_CHAN2)
	{
		llChannel = LL_DAC_CHANNEL_2 ;
		pFlagsChan = & pIoDesc->flagsChan [1] ;
	}
	else
	{
		return DAC_EARG ;
	}
	* pFlagsChan = pIoDesc->flags | flags ;

	// Compute DAC data register to use
	pDataReg = & pIoDesc->dataReg [channel-1u] ;
	switch (* pFlagsChan & (DAC_DUAL | DAC_BITS8))
	{
		case (DAC_BITS8):
			* pDataReg = (uint32_t)  ((channel == DAC_CHAN1) ? & pDac->DHR8R1 : & pDac->DHR8R2) ;
			break ;

		case (DAC_BITS12):
			* pDataReg = (uint32_t)  ((channel == DAC_CHAN1) ? & pDac->DHR12R1 : & pDac->DHR12R2) ;
			break ;

		case (DAC_DUAL | DAC_BITS8):
			* pDataReg = (uint32_t) & pDac->DHR8RD ;
			break ;

		default:	// (DAC_DUAL | DAC_BITS12)
			* pDataReg = (uint32_t) & pDac->DHR12RD ;
			break ;
	}
	if ((* pFlagsChan & DAC_ALIGNLEFT) != 0u)
	{
		* pDataReg += 4u ;
	}

	// Set the output buffer for the selected DAC channel (default is enabled)
	if ((flags & DAC_OUTBUFFDIS) != 0)
	{
		LL_DAC_SetOutputBuffer (pDac, llChannel, LL_DAC_OUTPUT_BUFFER_DISABLE) ;
	}

	// Set trigger source
	LL_DAC_SetTriggerSource (pDac, llChannel, trigSrc) ;

	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	Set the DAC callback for one independent channel or the dual channel
//	DMA and DAC interrupt callback

dacStatus_t 	dacSetCallback		(uint32_t channel, dmaCallback_t cb, uintptr_t arg)
{
	if (channel == DAC_CHAN1)
	{
		pIoDesc->cb[0]     = cb ;
		pIoDesc->cbArg [0] = arg ;
		if (pIoDesc->pDacDesc->dma1.dmaNum != DMA_NONE)
		{
			dmaSetCallback	(& pIoDesc->pDacDesc->dma1, cb, arg) ;
		}
	}
	else if (channel == DAC_CHAN2)
	{
		// Register the user callback
		pIoDesc->cb[1]     = cb ;
		pIoDesc->cbArg [1] = arg ;

		// Configure the DMA callback
		if (pIoDesc->pDacDesc->dma2.dmaNum != DMA_NONE)
		{
			dmaSetCallback	(& pIoDesc->pDacDesc->dma2, cb, arg) ;
		}
	}
	else
	{
		return DAC_EARG ;
	}

	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	Configure DMA
//	channel : one of DAC_CHAN1, DAC_CHAN2
//	flags   : one of DAC_DMA_xxx and optional DAC_DMA_NOIT
//	length  : count of items in DMA buffers

dacStatus_t 	dacDmaInit		(uint32_t channel, uint32_t flags, uint32_t length)
{
	const dmaDesc_t 	* pDmaDesc ;
	dma_t				* pDma ;
	uint32_t			llChannel ;
	uint32_t			streamNum ;
	DAC_TypeDef			* pDac = pIoDesc->pDacDesc->pDac ;
	uint32_t			reg ;
	uint32_t			* pFlagsChan ;

	if (channel == DAC_CHAN1)
	{
		llChannel = LL_DAC_CHANNEL_1 ;
		pDmaDesc  = & pIoDesc->pDacDesc->dma1 ;
		pFlagsChan = & pIoDesc->flagsChan [0] ;
	}
	else if (channel == DAC_CHAN2)
	{
		llChannel = LL_DAC_CHANNEL_2 ;
		pDmaDesc  = & pIoDesc->pDacDesc->dma2 ;
		pFlagsChan = & pIoDesc->flagsChan [1] ;
	}
	else
	{
		return DAC_EARG ;
	}

	if (pDmaDesc->dmaNum == DMA_NONE)
	{
		AA_ASSERT (0) ;
		return DAC_ESTATE ;		// No DMA descriptor
	}

	* pFlagsChan |= flags & DAC_DMA_MASK ;
	pDma		= DMA_GETDMA (pDmaDesc->dmaNum) ;
	streamNum	= DMA_GETSTREAM (pDmaDesc) ;

	// Enable DMA clock
	dmaEnableClock (pDma) ;

	// Reset stream
	dmaStreamResetRaw (pDma, streamNum) ;

	LL_DMA_SetChannelSelection ((DMA_TypeDef *) pDma, streamNum, DMA_GETCHANNEL (pDmaDesc) << 25);

	// Compute Peripheral data size
	switch (* pFlagsChan & (DAC_DUAL | DAC_BITS8))
	{
		case (DAC_BITS8):
			reg = LL_DMA_PDATAALIGN_BYTE ;
			break ;

		case (DAC_BITS12):
			reg = LL_DMA_PDATAALIGN_HALFWORD ;
			break ;

		case (DAC_DUAL | DAC_BITS8):
			reg = LL_DMA_PDATAALIGN_HALFWORD ;
			break ;

		default:	// (DAC_DUAL | DAC_BITS12)
			reg = LL_DMA_PDATAALIGN_WORD ;
			break ;
	}

	// Configure DMA
	LL_DMA_ConfigTransfer ((DMA_TypeDef *) pDma, streamNum,
							(((flags & DAC_DMA_CIRCULAR) != 0) ? LL_DMA_MODE_CIRCULAR : LL_DMA_MODE_NORMAL) |
							reg								  |		// LL_DMA_PDATAALIGN_xxx
							LL_DMA_DIRECTION_MEMORY_TO_PERIPH |
							LL_DMA_PERIPH_NOINCREMENT         |
							LL_DMA_MEMORY_INCREMENT           |
							LL_DMA_MDATAALIGN_HALFWORD) ;			// direct mode so: MSIZE is forced by hardware to PSIZE value


	// Set double buffer if required
	if ((flags & DAC_DMA_DOUBLE) != 0u)
	{
		LL_DMA_EnableDoubleBufferMode ((DMA_TypeDef *) pDma, streamNum) ;
	}

	// Set DMA transfer size
	LL_DMA_SetDataLength ((DMA_TypeDef *) pDma, streamNum, length) ;

	// Configure interrupts
	if ((flags & DAC_DMA_NOIT) == 0u)
	{
		//	Configure NVIC for DMA transfer complete/error interrupts
		dmaSetIrq (pDmaDesc) ;

		// Enable DMA transfer interrupt + transfer error
		LL_DMA_EnableIT_TE ((DMA_TypeDef *) pDma, streamNum) ;
		LL_DMA_EnableIT_TC ((DMA_TypeDef *) pDma, streamNum) ;
		if ((flags & DAC_DMA_CIRCULAR) != 0)
		{
			LL_DMA_EnableIT_HT ((DMA_TypeDef *) pDma, streamNum) ;
		}

		// Configure NVIC to enable DAC interruptions
		NVIC_SetPriority (TIM6_DAC_IRQn, DAC_INT_PRIORITY) ;
		NVIC_EnableIRQ   (TIM6_DAC_IRQn) ;

		// Enable DAC channel underrun interrupt
		if (flags & DAC_CHAN1)
		{
			LL_DAC_EnableIT_DMAUDR1 (pDac) ;
		}
		else
		{
			LL_DAC_EnableIT_DMAUDR2 (pDac) ;
		}
	}
	else
	{
		// Disable DMA interrupts
		LL_DMA_DisableIT_TE ((DMA_TypeDef *) pDma, streamNum) ;
		LL_DMA_DisableIT_TC ((DMA_TypeDef *) pDma, streamNum) ;

		if (flags & DAC_CHAN1)
		{
			LL_DAC_DisableIT_DMAUDR1 (pDac) ;
		}
		else
		{
			LL_DAC_DisableIT_DMAUDR2 (pDac) ;
		}
	}

	// Enable DAC channel DMA request
	LL_DAC_EnableDMAReq (pDac, llChannel);

	 return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	Set DMA buffers
//	If DAC_DMA_CIRCULAR or DAC_DMA_SINGLE only buffer 1 is used. To be called before dacStart()
//	If DAC_DMA_DOUBLE and called before dacStart(): buffer 1 and buffer 2 are used
//	If DAC_DMA_DOUBLE and called after dacStart(): only buffer 1 is used

dacStatus_t 	dacDmaBuffer	(uint32_t channel, const void * buffer1, const void * buffer2)
{
	const dmaDesc_t 	* pDmaDesc ;
	dma_t				* pDma ;
	uint32_t			streamNum ;
	DAC_TypeDef			* pDac = pIoDesc->pDacDesc->pDac ;
	uint32_t			ix ;

	// Find the DMA configuration according to the channel
	if (channel == DAC_CHAN1)
	{
		pDmaDesc = & pIoDesc->pDacDesc->dma1 ;
		ix		 = 0 ;
	}
	else
	{
		pDmaDesc = & pIoDesc->pDacDesc->dma2 ;
		ix		 = 1 ;
	}
	if (pDmaDesc->dmaNum == DMA_NONE)
	{
		AA_ASSERT (0) ;
		return DAC_ESTATE ;
	}
	pDma		= DMA_GETDMA (pDmaDesc->dmaNum) ;
	streamNum	= DMA_GETSTREAM (pDmaDesc) ;

	// Set DMA addresses
	if ((pDac->CR & DMA_SxCR_EN) == 0)
	{
		// DAC not enabled: initialize DMA buffers
		LL_DMA_SetMemoryAddress ((DMA_TypeDef *) pDma, streamNum, (uint32_t) buffer1) ;
		LL_DMA_SetPeriphAddress ((DMA_TypeDef *) pDma, streamNum, pIoDesc->dataReg [ix]) ;

		if ((pIoDesc->flagsChan [ix] & DAC_DMA_DOUBLE) != 0)
		{
			// Also set second buffer address
			LL_DMA_SetMemory1Address ((DMA_TypeDef *) pDma, streamNum, (uint32_t) buffer2) ;
		}
	}
	else
	{
		// DAC enabled : update memory DMA buffer not in use (double buffer only)
		if (LL_DMA_GetCurrentTargetMem ((DMA_TypeDef *) pDma, streamNum) == 0)
		{
			LL_DMA_SetMemory1Address ((DMA_TypeDef *) pDma, streamNum, (uint32_t) buffer1) ;
		}
		else
		{
			LL_DMA_SetMemoryAddress ((DMA_TypeDef *) pDma, streamNum, (uint32_t) buffer1) ;
		}
	}

	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	Start DAC generation on specified channel

dacStatus_t 	dacStart		(uint32_t channel)
{
	const dmaDesc_t 	* pDmaDesc ;
	dma_t				* pDma ;
	uint32_t			streamNum ;
	DAC_TypeDef			* pDac = pIoDesc->pDacDesc->pDac ;
	uint32_t			llChannel ;
	__IO uint32_t		wait_loop_index ;

	// Find the DMA configuration according to the channel
	if (channel == DAC_CHAN1)
	{
		llChannel = LL_DAC_CHANNEL_1 ;
		pDmaDesc = & pIoDesc->pDacDesc->dma1 ;
	}
	else
	{
		llChannel = LL_DAC_CHANNEL_2 ;
		pDmaDesc = & pIoDesc->pDacDesc->dma2 ;
	}

	// Enable the DMA transfer
	if (pDmaDesc->dmaNum != DMA_NONE)
	{
		pDma		= DMA_GETDMA (pDmaDesc->dmaNum) ;
		streamNum	= DMA_GETSTREAM (pDmaDesc) ;
		LL_DMA_EnableStream ((DMA_TypeDef *) pDma, streamNum) ;
	}

	// Enable DAC channel
	LL_DAC_Enable (pDac, llChannel) ;
	wait_loop_index = ((LL_DAC_DELAY_STARTUP_VOLTAGE_SETTLING_US * (SystemCoreClock / (100000 * 2))) / 10) ;
	while(wait_loop_index != 0)
	{
		wait_loop_index-- ;
	}

	// Enable DAC channel trigger
	LL_DAC_EnableTrigger (pDac, llChannel) ;

	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	Stop DAC generation on specified channel

dacStatus_t 	dacStop		(uint32_t channel)
{
	const dmaDesc_t 	* pDmaDesc ;
	DAC_TypeDef			* pDac = pIoDesc->pDacDesc->pDac ;
	uint32_t			llChannel ;

	// Find the DMA configuration according to the channel
	if (channel == DAC_CHAN1)
	{
		llChannel = LL_DAC_CHANNEL_1 ;
		pDmaDesc = & pIoDesc->pDacDesc->dma1 ;
	}
	else
	{
		llChannel = LL_DAC_CHANNEL_2 ;
		pDmaDesc = & pIoDesc->pDacDesc->dma2 ;
	}

	LL_DAC_DisableTrigger	(pDac, llChannel) ;
	LL_DAC_Disable 			(pDac, llChannel) ;
	if(pDmaDesc->dmaNum != DMA_NONE)
	{
		dmaStreamReset (pDmaDesc) ;
	}
	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	Set static output value

dacStatus_t 	dacOuput		(uint32_t channel, uint32_t value)
{
	if (channel == DAC_CHAN1)
	{
		* (uint32_t *) pIoDesc->dataReg [0] = value ;
	}
	else if (channel == DAC_CHAN2)
	{
		* (uint32_t *) pIoDesc->dataReg [1] = value ;
	}
	else
	{
		return DAC_EARG ;
	}
	return DAC_ENONE ;
}

//----------------------------------------------------------------------
// Triggers DAC conversion by software

dacStatus_t 	dacSwTrigger	(uint32_t channel)
{
//	LL_DAC_TrigSWConversion (pIoDesc->pDacDesc->pDac, (channel == DAC_CHAN1) ? LL_DAC_CHANNEL_1 : LL_DAC_CHANNEL_2) ;
	LL_DAC_TrigSWConversion (pIoDesc->pDacDesc->pDac, channel) ;

    return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	DAC interrupt handler, called from timer interrupt
//	Under run interrupt

void	dac1BasicIrqHandler (void) ;
void	dac1BasicIrqHandler (void)
{
	// Check whether DAC channel1 underrun caused the DAC interrupt
	if (LL_DAC_IsActiveFlag_DMAUDR1 (DAC1) != 0u)
	{
		// Clear flag DAC channel1 underrun
		LL_DAC_ClearFlag_DMAUDR1 (DAC1) ;

	    // Call interrupt callback function
		if (pIoDesc->cb [0] != NULL)
		{
			(* pIoDesc->cb [0]) (DAC_EV_UNDR, DAC_CHAN1, pIoDesc->cbArg [0]) ;
		}
	}

	// Check whether DAC channel2 underrun caused the DAC interrupt
	if (LL_DAC_IsActiveFlag_DMAUDR2 (DAC1) != 0u)
	{
		// Clear flag DAC channel2 underrun
		LL_DAC_ClearFlag_DMAUDR2 (DAC1) ;

	    // Call interrupt callback function
		if (pIoDesc->cb [1] != NULL)
		{
			(* pIoDesc->cb [1]) (DAC_EV_UNDR, DAC_CHAN2, pIoDesc->cbArg [1]) ;
		}
	}
}

//----------------------------------------------------------------------
#endif	// defined DAC  ||  defined DAC1
