/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	i2cbasic.c	Master and Slave basic routines:
				Master and Slave use interrupt mode,
				7 bits address only (no 10 bits address)
				Slave have only one address, no general call
				Standard (<100 KHz) or Fast Mode (<400 KHz)

	When		Who	What
	01/03/18	ac	Creation
	22/12/21	ac	Rework again

	This code is a rework of STMicroelectronics application note AN2824

	BEWARE: From the STM32F4 errata sheet:
	Do not use repeated start in standard mode
	when the bit rate is between 88 kHz and 100 kHz

	To use ReSTART, the slave receive buffer size must be > master data size + 2.
	Otherwise the slave will send a NACK which will disrupt the sequence.

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2023 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aakernel.h"

#include	"i2cbasic.h"
#include	"rccbasic.h"

#include	"stm32f4xx_ll_i2c.h"

//----------------------------------------------------------------------
//	 For debug purpose

#define	DBG_I2C					1		// 1 to display data on SWO

#if (DBG_I2C != 0)
	#include	"itm.h"
	#pragma GCC diagnostic push
	#pragma GCC diagnostic ignored "-Wmultichar"
#endif

//----------------------------------------------------------------------

#define I2C_DIRECTION_TX 		0
#define I2C_DIRECTION_RX 		1

// http://fastbitlab.com/stm32-i2c-lecture-12-i2c-serial-clock-settings-with-explanations/
// https://stackoverflow.com/questions/31384447/i2c-duty-cycle-significance
//#define I2C_DUTYCYCLE			LL_I2C_DUTYCYCLE_2		// Fast Mode Thigh/Tlow=2
#define I2C_DUTYCYCLE			LL_I2C_DUTYCYCLE_16_9	// Fast Mode Thigh/Tlow=1.8 => up to 400KHz

//----------------------------------------------------------------------
// Transaction descriptor, useful for i2c functions using interrupts

typedef struct
{
	I2C_TypeDef		* pI2c ;			// I2C1, I2C2...
	uint8_t			* pBuffer ;
	uint16_t		size ;
	uint16_t		idx ;
	uint32_t		flags ;				// Configuration and status flags
	uint32_t		slaveAddress ;		// Own slave address
	i2cStatus_t		status ;			// Result of the operation
	aaTaskId_t		taskId ;			// The task to signal at end of operation
	aaSignal_t		signal ;			// The signal to use

} i2cIo_t ;

// For i2cIo_t.flags
#define	I2C_FLAG_DIRMASK		0x00000001u
#define	I2C_FLAG_TX				0x00000000u
#define	I2C_FLAG_RX				0x00000001u

#define	I2C_FLAG_MODEMASK		0x00000010u
#define	I2C_FLAG_MASTER			0x00000000u
#define	I2C_FLAG_SLAVE			0x00000010u

#define	I2C_FLAG_SEQ			0x00000000u	// Master sequential mode:use Xfer
#define	I2C_FLAG_AUTO			0x00000100u	// Master auto mode: i2cMasterwrite(), i2cMasteRead()

#define	I2C_FLAG_LASTMASK		0x00001000u	// Mask of LAST xfer value
#define	I2C_FLAG_NEXT			0x00000000u	// Not last xfer
#define	I2C_FLAG_LAST			0x00001000u	// Last xfer before STOP or RESTART

#define	I2C_STATE_MASK			0x000F0000u	// Mask of mode value
#define	I2C_STATE_IDLE			0x00010000u	// Interface not initialized
#define	I2C_STATE_ADDR			0x00020000u	// In master mode, waiting for end of start condition and slave address transmission
#define	I2C_STATE_TX			0x00030000u	// Interface xfer TX
#define	I2C_STATE_TXWAIT		0x00040000u	// Interface waiting for new data or stop or restart condition
#define	I2C_STATE_RX			0x00050000u	// Interface xfer RX
#define	I2C_STATE_RXWAIT		0x00060000u	// Interface waiting for new data or stop or restart condition

// To set i2cIo_t.flags
__STATIC_INLINE void setFlagState (i2cIo_t * pI2cIo, uint32_t state)
{
	pI2cIo->flags = (pI2cIo->flags & ~I2C_STATE_MASK) | (state & I2C_STATE_MASK) ;
}

__STATIC_INLINE void setFlagAuto (i2cIo_t * pI2cIo, uint32_t autoFlag)
{
	pI2cIo->flags = (pI2cIo->flags & ~I2C_FLAG_AUTO) | (autoFlag & I2C_FLAG_AUTO) ;
}

__STATIC_INLINE void setFlagDir (i2cIo_t * pI2cIo, uint32_t dir)
{
	pI2cIo->flags = (pI2cIo->flags & ~I2C_FLAG_DIRMASK) | (dir & I2C_FLAG_DIRMASK) ;
}

__STATIC_INLINE void setFlagMode (i2cIo_t * pI2cIo, uint32_t mode)
{
	pI2cIo->flags = (pI2cIo->flags & ~I2C_FLAG_MODEMASK) | (mode & I2C_FLAG_MODEMASK) ;
}

__STATIC_INLINE void setFlagLast (i2cIo_t * pI2cIo, uint32_t nextOrLast)
{
	pI2cIo->flags = (pI2cIo->flags & ~I2C_FLAG_LASTMASK) | (nextOrLast & I2C_FLAG_LASTMASK) ;
}


// To get i2cIo_t.flags
__STATIC_INLINE uint32_t getFlagState (i2cIo_t * pI2cIo)
{
	return pI2cIo->flags & I2C_STATE_MASK ;
}

__STATIC_INLINE uint32_t getFlagAuto (i2cIo_t * pI2cIo)
{
	return pI2cIo->flags & I2C_FLAG_AUTO ;
}

__STATIC_INLINE uint32_t getFlagDir (i2cIo_t * pI2cIo)
{
	return pI2cIo->flags & I2C_FLAG_DIRMASK ;
}

__STATIC_INLINE uint32_t getFlagMode (i2cIo_t * pI2cIo)
{
	return pI2cIo->flags & I2C_FLAG_MODEMASK ;
}

__STATIC_INLINE uint32_t getFlagLast (i2cIo_t * pI2cIo)
{
	return pI2cIo->flags & I2C_FLAG_LASTMASK ;
}

//----------------------------------------------------------------------

#if (WITH_I2C_X == 1)

	#if (defined STM32F4)
		#include "i2cF4xx.h"		//	I2C descriptors for STM32F4xx
	#endif

	STATIC const uint32_t		i2cArrayCount = sizeof (i2cArray) / sizeof (i2cDesc_t) ;

#endif

//----------------------------------------------------------------------
//	One I/O descriptor for each physical I2C peripheral

	#if (WITH_I2C_1 == 1)
		i2cIo_t		i2cIo1 ;
	#endif	// WITH_I2C_1

	#if (WITH_I2C_2 == 1)
		i2cIo_t		i2cIo2 ;
	#endif	// WITH_I2C_2

	#if (WITH_I2C_3 == 1)
		i2cIo_t		i2cIo3 ;
	#endif	// WITH_I2C_3

	#if (WITH_I2C_4 == 1)
		i2cIo_t		i2cIo4 ;
	#endif	// WITH_I2C_4

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	Get the I/O descriptor address for an I2C peripheral

static	i2cIo_t *	i2cGetIoDesc (I2C_TypeDef * pI2C) ;

static	i2cIo_t *	i2cGetIoDesc (I2C_TypeDef * pI2C)
{
	#if (WITH_I2C_X == 0)
	{
		(void) pI2C ;	// To avoid compiler warning when none I2C is supported
	}
	#endif

	#if (WITH_I2C_1 == 1)
	{
		if (pI2C == I2C1)
		{
			return & i2cIo1 ;
		}
	}
	#endif	// WITH_I2C_1

	#if (WITH_I2C_2 == 1)
	{
		if (pI2C == I2C2)
		{
			return & i2cIo2 ;
		}
	}
	#endif	// WITH_I2C_2

	#if (WITH_I2C_3 == 1)
	{
		if (pI2C == I2C3)
		{
			return & i2cIo3 ;
		}
	}
	#endif	// WITH_I2C_3

	#if (WITH_I2C_4 == 1)
	{
		if (pI2C == I2C4)
		{
			return & i2cIo4 ;
		}
	}
	#endif // WITH_I2C_4

	AA_ASSERT (0) ;
	return NULL ;
}

//----------------------------------------------------------------------
//	Get an I2C descriptor from its identifier

const i2cDesc_t *	i2cGetDesc	(uint32_t i2cId)
{
	#if (WITH_I2C_X == 1)
	{
		uint32_t	ii ;

		for (ii = 0u ; ii < i2cArrayCount ; ii++)
		{
			if (i2cArray [ii].id == i2cId)
			{
				return & i2cArray [ii] ;
			}
		}
	}
	#else
	{
		(void) i2cId ;		// To avoid compiler warning when none I2C is supported
	}
	#endif

	return NULL ;	// Identifier not found
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
// 	Configure the I2C peripheral using the descriptor, for master or slave I2C mode
//	Give the I2C handle to use as parameter for the other functions

//	Clock stretching enabled
//	7 bits address
//	Own address: 0
//	No own address 2
//	No general call

i2cStatus_t	i2cConfigure (const i2cDesc_t * pI2cDesc, uint32_t speed, aaSignal_t signal, i2cHandle_t * pHdl)
{
	i2cIo_t		* pI2cIo ;
	uint32_t	i2cClock ;

	pI2cIo = i2cGetIoDesc (pI2cDesc->pI2c) ;
	if (pI2cIo == NULL)
	{
		return I2C_EFAIL ;
	}
	aaMEMSET (pI2cIo, 0, sizeof (i2cIo_t)) ;
	pI2cIo->pI2c = pI2cDesc->pI2c ;

	gpioConfigurePin (& pI2cDesc->scl) ;
	gpioConfigurePin (& pI2cDesc->sda) ;

	// Enable the I2C peripheral clock, and reset the peripheral
	rccEnableClock (pI2cDesc->pI2c, 1u << pI2cDesc->periphEnPos) ;
	rccResetPeriph (pI2cDesc->pI2c, 1u << pI2cDesc->periphRstPos) ;

	// Configure Event IT:
	// - Set priority for I2Cx_EV_IRQn
	// - Enable I2Cx_EV_IRQn
	NVIC_SetPriority (pI2cDesc->evIrqn, pI2cDesc->irqPriority);
	NVIC_EnableIRQ   (pI2cDesc->evIrqn) ;

	// Configure Error IT:
	// - Set priority for I2Cx_ER_IRQn
	// - Enable I2Cx_ER_IRQn
	NVIC_SetPriority (pI2cDesc->erIrqn, pI2cDesc->irqPriority);
	NVIC_EnableIRQ   (pI2cDesc->erIrqn) ;

	// Retrieve Clock frequencies
	i2cClock = rccGetClockFrequency (pI2cDesc->pI2c) ;

	// Configure the SCL Clock Speed
	LL_I2C_ConfigSpeed (pI2cDesc->pI2c, i2cClock, speed, I2C_DUTYCYCLE) ;

	// Configure the Own Address1
	// Reset Values :
	//    - OwnAddress1 is 0x00
	//    - OwnAddrSize is LL_I2C_OWNADDRESS1_7BIT
	//LL_I2C_SetOwnAddress1(pI2cDesc->pI2c, 0x00, LL_I2C_OWNADDRESS1_7BIT);

	// Enable Clock stretching
	// Reset Value is Clock stretching enabled
	//LL_I2C_EnableClockStretching (pI2cDesc->pI2c) ;

	// Enable General Call
	// Reset Value is General Call disabled
	//LL_I2C_EnableGeneralCall (pI2cDesc->pI2c) ;

	// Configure the 7bits Own Address2
	// Reset Values :
	//    - OwnAddress2 is 0x00
	//    - Own Address2 is disabled
	//LL_I2C_SetOwnAddress2     (pI2cDesc->pI2c, ownAddress2);
	//LL_I2C_DisableOwnAddress2 (pI2cDesc->pI2c) ;

	// Enable Peripheral in I2C mode
	// Reset Value is I2C mode
	//LL_I2C_SetMode (pI2cDesc->pI2c, LL_I2C_MODE_I2C) ;

	pI2cIo->signal	= signal ;
	pI2cIo->status	= I2C_EOK ;
	pI2cIo->taskId	= AA_INVALID_TASK ;
	setFlagState (pI2cIo, I2C_STATE_IDLE) ;

	* pHdl = (i2cHandle_t) pI2cIo ;
	return I2C_ENONE ;
}

//----------------------------------------------------------------------

void i2cDeactivate (i2cHandle_t hdl)
{
	LL_I2C_Disable (((i2cIo_t *) hdl)->pI2c) ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	Check slave availability. Master should be configured before
//	Doesn't send or receive any data byte to/from the device
//	Returns I2C_EOK if slave answered ACK
//	Returns I2C_EAF if slave doesn't answered

i2cStatus_t		i2cMasterProbeDevice (i2cHandle_t hdl, uint32_t slaveAddress)
{
	i2cStatus_t	status ;

	status = i2cMasterStart (hdl, slaveAddress, I2C_TX) ;	// TX mandatory
	i2cMasterStop (hdl) ;
	return status ;
}

//----------------------------------------------------------------------
//	Send START or ReSTART condition and slave 7 bits address
//	Direction : I2C_TX or I2C_RX
//	Returns I2C_EOK in case of success, I2C_EAF if the slave doesn't acknowledge the address,
//	else returns another error value

//	At the end:
//		on TX the ADDR flag is cleared
//		on RX the ADDR flag is not cleared. This retain the slave to send its 1st byte

i2cStatus_t i2cMasterStart (i2cHandle_t hdl, uint32_t slaveAddress, uint32_t direction)
{
	i2cIo_t			* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef		* pI2c = pI2cIo->pI2c ;

	if (getFlagMode (pI2cIo) != I2C_FLAG_MASTER)
	{
		AA_ASSERT (0) ;
		pI2cIo->status = I2C_ESTATE ;
		return pI2cIo->status ;
	}

	// Activate  I2C
	pI2c->CR1 |= I2C_CR1_PE ;

	// Clear previous AF or ARLO
	pI2c->SR1 &= ~(I2C_SR1_AF | I2C_SR1_ARLO | I2C_SR1_TIMEOUT) ;

	// Enable ACK
	pI2c->CR1 |= I2C_CR1_ACK ;

	// Set slave address
	if ((direction & I2C_TX) == I2C_TX)
	{
		setFlagDir (pI2cIo, I2C_FLAG_TX) ;
		pI2cIo->slaveAddress = slaveAddress & ~I2C_OAR1_ADD0 ;	// Reset the address bit 0 for write
	}
	else
	{
		setFlagDir (pI2cIo, I2C_FLAG_RX) ;
		pI2cIo->slaveAddress = slaveAddress | I2C_OAR1_ADD0 ;	// Set the address bit 0 for read
	}

	setFlagState (pI2cIo, I2C_STATE_ADDR) ;
	pI2cIo->status	= I2C_EBUSY ;
	pI2cIo->taskId	= aaTaskSelfId () ;
	aaSignalClear (pI2cIo->taskId, pI2cIo->signal) ;

	// Send START condition
if (DBG_I2C != 0) swoSend16 ('s-', 0) ;
	pI2c->CR1 |= I2C_CR1_START ;
	// Wait until START is done before enabling interrupts
	// Else we can get a BTF interrupt before the start...
	// There is nothing to clear the BTF
	while ((pI2c->CR1 & I2C_CR1_START) != 0u)
	{
	}

	// Enable ERR and EVT IT (not BUFF IT)
	pI2c->CR2 |= I2C_CR2_ITERREN | I2C_CR2_ITEVTEN ;

	aaSignalWait (pI2cIo->signal, NULL, AA_SIGNAL_OR, AA_INFINITE) ;
	return pI2cIo->status ;
}

//----------------------------------------------------------------------
// Send STOP condition

void		i2cMasterStop (i2cHandle_t hdl)
{
	i2cIo_t			* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef		* pI2c = pI2cIo->pI2c ;

	// Send STOP condition only if I2C is busy
	if ((pI2c->SR2 & I2C_SR2_BUSY) == I2C_SR2_BUSY)
	{
if (DBG_I2C != 0) swoSend16 ('p-', 0) ;
		pI2c->CR1 |= I2C_CR1_STOP ;
		// Make sure that the STOP bit is cleared by hardware
		while ((pI2c->CR1 & I2C_CR1_STOP) == I2C_CR1_STOP)
		{
		}
		setFlagState (pI2cIo, I2C_STATE_IDLE) ;
	}
}

//----------------------------------------------------------------------
//	The master sends or receives data as requested with i2cMasterStart()
//	Non blocking mode (I2C_NOBLOCK flag in flags): returns after initiating the transfer
//		At the end of the transfer the result is memorized and can be checked with i2cResult()
//	Blocking mode (default): returns after the end of the transfer
//	Flags must contain one of I2C_NEXT or I2C_LAST

// On TX, if result is I2C_EAF then
//		Must send STOP or ReSTART
//		Must check: the message may be sent correctly if size == idx

// On RX
//		The minimum size of the last block is 2 bytes
//		If the same block is the first and the last, its size can be 1 byte

i2cStatus_t		i2cMasterXfer	(i2cHandle_t hdl, uint8_t * pBuffer, uint32_t numByte, uint32_t nextOrLast)
{
	i2cIo_t			* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef		* pI2c = pI2cIo->pI2c ;

	// Check if i2cMasterStart() already done
	{
		uint32_t temp = getFlagState (pI2cIo) ;
		if ((getFlagMode (pI2cIo) != I2C_FLAG_MASTER)   ||
			(temp != I2C_STATE_TXWAIT   &&
			 temp != I2C_STATE_ADDR     &&
			 temp != I2C_STATE_RXWAIT))
		{
			pI2cIo->status	= I2C_ESTATE ;
			pI2cIo->idx		= (int16_t) 0u ;
			return pI2cIo->status ;
		}
	}

	if (numByte == 0u)
	{
		pI2cIo->status	= I2C_EFAIL ;
		pI2cIo->idx		= (int16_t) 0u ;
		return I2C_EFAIL ;
	}

	// Initialize IO buffer
	pI2cIo->pBuffer = pBuffer ;
	pI2cIo->size	= numByte ;
	pI2cIo->idx		= (int16_t) 0u ;
	pI2cIo->status	= I2C_EBUSY ;
	pI2cIo->taskId	= aaTaskSelfId () ;
	aaSignalClear (pI2cIo->taskId, pI2cIo->signal) ;

	setFlagLast (pI2cIo, (nextOrLast & I2C_NEXT) == I2C_NEXT ? I2C_FLAG_NEXT : I2C_FLAG_LAST) ;

	// Set state
	if (getFlagDir (pI2cIo) == I2C_FLAG_TX)
	{
		// TX direction.
		setFlagState (pI2cIo, I2C_STATE_TX) ;
	}
	else
	{
		// RX direction.
		// Ti2cMasterStart() or previous i2cMasterXfer() has already done it
	}

	// Enable ERR, EVT and BUF IT
if (DBG_I2C != 0) swoSend16 ('X-', 0) ;
	pI2c->CR2 |= I2C_CR2_ITERREN | I2C_CR2_ITEVTEN | I2C_CR2_ITBUFEN ;

	if ((nextOrLast & I2C_NOBLOCK) == I2C_NOBLOCK)
	{
		return I2C_ENONE ;
	}
	// Wait for the transfer end
	aaSignalWait (pI2cIo->signal, NULL, AA_SIGNAL_OR, AA_INFINITE) ;
	return pI2cIo->status ;
}

//----------------------------------------------------------------------
//	Sends data to slave by performing all phases: START + data + STOP
//	Returns I2C_EOK if all bytes sent
//	Returns I2C_EAF if the slave has acknowledged before the end of the data

i2cStatus_t i2cMasterWrite (i2cHandle_t hdl, uint32_t slaveAddress, uint8_t * pBuffer, uint32_t numByteToWrite)
{
	i2cIo_t			* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef		* pI2c = pI2cIo->pI2c ;

	// Initialize IO descriptor
	pI2cIo->pBuffer 		= pBuffer ;
	pI2cIo->size			= numByteToWrite ;
	pI2cIo->slaveAddress	= slaveAddress & ~1u ;	// Set the address bit 0 for slave to receive
	pI2cIo->idx				= (int16_t) 0u ;
	pI2cIo->status			= I2C_EBUSY ;
	pI2cIo->taskId			= aaTaskSelfId () ;
	aaSignalClear (pI2cIo->taskId, pI2cIo->signal) ;

	// Set the I2C direction to Transmission
	setFlagDir   (pI2cIo, I2C_FLAG_TX) ;
	setFlagLast  (pI2cIo, I2C_FLAG_LAST) ;
	setFlagState (pI2cIo, I2C_STATE_ADDR) ;
	setFlagAuto  (pI2cIo, I2C_FLAG_AUTO) ;

	// Enable I2c
	pI2c->CR1 |= I2C_CR1_PE ;

	// Send START condition
	pI2c->CR1 |= I2C_CR1_START ;

	// Enable ERR and EVT IT (and not BUF IT)
	pI2c->CR2 |= I2C_CR2_ITERREN | I2C_CR2_ITEVTEN ;

	aaSignalWait (pI2cIo->signal, NULL, AA_SIGNAL_OR, AA_INFINITE) ;

	// Wait while busy (wait end of STOP)
	while ((pI2c->SR2 & I2C_SR2_BUSY) == I2C_SR2_BUSY)
	{
	}
	setFlagAuto  (pI2cIo, I2C_FLAG_SEQ) ;
	return pI2cIo->status ;
}


//----------------------------------------------------------------------
//	Receive data from slave by performing all phases: START + data + STOP

i2cStatus_t	i2cMasterRead (i2cHandle_t hdl, uint32_t slaveAddress, uint8_t * pBuffer, uint32_t numByteToRead)
{
	i2cIo_t			* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef		* pI2c = pI2cIo->pI2c ;

	// Initialize IO descriptor
	pI2cIo->pBuffer 		= pBuffer ;
	pI2cIo->size			= numByteToRead ;
	pI2cIo->slaveAddress	= slaveAddress | 1u ;	// Set the address bit 0 for slave to transmit
	pI2cIo->idx				= (int16_t) 0u ;
	pI2cIo->status			= I2C_EBUSY ;
	pI2cIo->taskId			= aaTaskSelfId () ;
	aaSignalClear (pI2cIo->taskId, pI2cIo->signal) ;

	// Set the I2C direction to reception
	setFlagDir  (pI2cIo, I2C_FLAG_RX) ;
	setFlagLast  (pI2cIo, I2C_FLAG_LAST) ;
	setFlagState (pI2cIo, I2C_STATE_ADDR) ;
	setFlagAuto  (pI2cIo, I2C_FLAG_AUTO) ;

	// Enable I2c
	pI2c->CR1 |= I2C_CR1_PE ;

	// Enable Acknowledge
	pI2c->CR1  |= I2C_CR1_ACK ;

	// Send START condition
	pI2c->CR1 |= I2C_CR1_START ;

	// Enable ERR and EVT IT (and not BUF IT)
	pI2c->CR2 |= I2C_CR2_ITERREN | I2C_CR2_ITEVTEN ;

	aaSignalWait (pI2cIo->signal, NULL, AA_SIGNAL_OR, AA_INFINITE) ;

	// Wait while busy (wait end of STOP)
 	while ((pI2c->SR2 & I2C_SR2_BUSY) == I2C_SR2_BUSY)
	{
	}
	setFlagAuto  (pI2cIo, I2C_FLAG_SEQ) ;
	return pI2cIo->status ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	Configure the slave own address
//	The slave is set in non addressable state

void	i2cSlaveInit (i2cHandle_t hdl, uint32_t ownAddress1)
{
	i2cIo_t			* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef		* pI2c = pI2cIo->pI2c ;

	// Configure the Own Address1 :
	LL_I2C_SetOwnAddress1 (pI2c, ownAddress1, LL_I2C_OWNADDRESS1_7BIT) ;

	setFlagMode (pI2cIo, I2C_FLAG_SLAVE) ;
}

//----------------------------------------------------------------------
//	Set the slave in addressable state
//	Blocking function until the slave is addressed by the master
//	The return value is I2C_ESLAVERX or I2C_ESLAVETX, or an error value
//	On return can call i2cResult() to know if the slave is addressed as TX or RX

i2cStatus_t 	i2cSlaveListen		(i2cHandle_t hdl, uint32_t flags)
{
	i2cIo_t		* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef	* pI2c = pI2cIo->pI2c ;
	(void) flags ;

	if (getFlagMode (pI2cIo) != I2C_FLAG_SLAVE)
	{
		pI2cIo->status = I2C_ESTATE ;
		return pI2cIo->status ;
	}

	// Initialize IO descriptor
	pI2cIo->status	= I2C_EBUSY ;
	pI2cIo->taskId	= aaTaskSelfId () ;
	aaSignalClear (pI2cIo->taskId, pI2cIo->signal) ;

	// Set slave in addressable state
	setFlagState (pI2cIo, I2C_STATE_ADDR) ;

	// Enable IT needed for ADDR / STOPF ITs, Do not enable Buffer IT (TXE and RXNE)
if (DBG_I2C != 0) swoSend16 ('L/', 0) ;
	pI2c->CR2 |= I2C_CR2_ITEVTEN | I2C_CR2_ITERREN ;

	pI2c->CR1 |= I2C_CR1_PE ;		// Activate slave I2C

	// Prepare acknowledge for Slave address reception
	pI2c->CR1 |= I2C_CR1_ACK ;	// To be done only if PE is set

	aaSignalWait (pI2cIo->signal, NULL, AA_SIGNAL_OR, AA_INFINITE) ;
	return pI2cIo->status ;
}

//----------------------------------------------------------------------
//	Set the slave in transfer state: wait a transfer initiated by the master
//	Non blocking mode (I2C_NOBLOCK flag in flags): returns after initiating the transfer
//		At the end of the transfer the result is memorized and can be checked with i2cResult()
//	Blocking mode (default): returns after the end of the transfer
//	Flags must contain one of I2C_NEXT or I2C_LAST

i2cStatus_t		i2cSlaveXfer		(i2cHandle_t hdl, uint8_t * pBuffer, uint32_t numByte, uint32_t nextOrLast)
{
	i2cIo_t		* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef	* pI2c = pI2cIo->pI2c ;
	uint32_t	state = getFlagState (pI2cIo) ;

	if (state != I2C_STATE_TXWAIT  &&  state != I2C_STATE_RXWAIT)
	{
		pI2cIo->status = I2C_ESTATE ;
		return pI2cIo->status ;
	}

	if (numByte == 0u)
	{
		pI2cIo->status = I2C_EFAIL ;
		return pI2cIo->status ;
	}

	pI2cIo->status	= I2C_EBUSY ;
	pI2cIo->pBuffer = pBuffer ;
	pI2cIo->size	= (int16_t) numByte ;
	pI2cIo->idx		= (int16_t) 0u ;		// Count of exchanged bytes
	pI2cIo->taskId	= aaTaskSelfId () ;
	aaSignalClear (pI2cIo->taskId, pI2cIo->signal) ;

	setFlagLast (pI2cIo, ((nextOrLast & I2C_NEXT) == I2C_NEXT) ? I2C_FLAG_NEXT : I2C_FLAG_LAST) ;

	if (getFlagState (pI2cIo) == I2C_STATE_TXWAIT)
	{
		// Addressed as transmitter
		setFlagState (pI2cIo, I2C_STATE_TX) ;
	}
	else
	{
		// Addressed as receiver
		setFlagState (pI2cIo, I2C_STATE_RX) ;
		if (numByte == 1u)
		{
			// Only 1 byte to receive, clear ACK
			pI2c->CR1 &= ~I2C_CR1_ACK ;
		}
	}
	// Enable data interrupt
if (DBG_I2C != 0) swoSend16 ('X/', 0) ;
	pI2c->CR2 |= I2C_CR2_ITEVTEN | I2C_CR2_ITBUFEN | I2C_CR2_ITERREN ;

	if ((nextOrLast & I2C_NOBLOCK) == I2C_NOBLOCK)
	{
		return I2C_ENONE ;
	}
	aaSignalWait (pI2cIo->signal, NULL, AA_SIGNAL_OR, AA_INFINITE) ;
	return pI2cIo->status ;
}

//----------------------------------------------------------------------
//	Returns:
//		I2C_EBUSY	if the operation is not ended
//		I2C_EOK		if the operation is ended without error
//		Another error value
//	At the end of a transfer the effective count of transfered bytes is copied to * pSize

i2cStatus_t		i2cResult		(i2cHandle_t hdl, uint32_t * pSize)
{
	i2cIo_t		* pI2cIo = (i2cIo_t *) hdl ;
	i2cStatus_t	result ;

	result = pI2cIo->status ;
	if (pSize != NULL)
	{
		if (result != I2C_EBUSY)
		{
			* pSize = pI2cIo->idx ;
		}
		else
		{
			* pSize = 0u ;
		}
	}
	return result ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	Interrupt handlers

//	End of phase: update internal state and signal the task

static	void	i2cEnd	(i2cIo_t * pI2cIo, i2cStatus_t status, uint32_t state)
{
	I2C_TypeDef		* pI2c = pI2cIo->pI2c ;

	// Disable all interrupt sources
	pI2c->CR2 &= ~(I2C_CR2_ITEVTEN | I2C_CR2_ITERREN | I2C_CR2_ITBUFEN) ;

	setFlagState (pI2cIo, state) ;
	pI2cIo->status = status ;

	// Signal the task
	if (pI2cIo->taskId != AA_INVALID_TASK)
	{
		aaSignalSend (pI2cIo->taskId, pI2cIo->signal) ;
	}
	pI2cIo->taskId = AA_INVALID_TASK ;
}

//----------------------------------------------------------------------

void I2Cx_EV_IRQHandler (I2C_TypeDef * pI2c, i2cIo_t * pI2cIo)
{
	__IO uint32_t SR1Register ;
	__IO uint32_t SR2Register ;

	aaIntEnter () ;

	// Read -ONLY- I2C SR1 status register
	SR2Register = 0u ;
	SR1Register = pI2c->SR1 ;

	if (getFlagMode (pI2cIo) == I2C_FLAG_SLAVE)
	{
		// Slave mode
		// If ADDR = 1: EV1
//if (DBG_I2C != 0) swoSend16 ('1/', 0) ;
		if ((SR1Register & I2C_SR1_ADDR) == I2C_SR1_ADDR)
		{
			if (getFlagState (pI2cIo) != I2C_STATE_ADDR)
			{
				// ReSTART received
				// End of block without clearing ADDR
				if ((getFlagState (pI2cIo) == I2C_STATE_TXWAIT))
				{
					i2cEnd (pI2cIo, (pI2cIo->idx == pI2cIo->size) ? I2C_EOK : I2C_EAF, I2C_STATE_IDLE) ;
				}
				else
				{
if (DBG_I2C != 0) swoSend16 ('2/', 0) ;
if (DBG_I2C != 0) swoSend8 ('0'+ ((pI2cIo->flags >> 16) & 0x0F), 0) ;
					i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_IDLE) ;
				}
			}
			else
			{
				// State I2C_STATE_ADDR
if (DBG_I2C != 0) swoSend16 ('3/', 0) ;
				SR2Register = pI2c->SR2 ;		// This clears the ADDR
				if ((SR2Register & I2C_SR2_TRA) != 0u)
				{
					i2cEnd (pI2cIo, I2C_ESLAVETX, I2C_STATE_TXWAIT) ;	// The slave is sender
				}
				else
				{
					i2cEnd (pI2cIo, I2C_ESLAVERX, I2C_STATE_RXWAIT) ;	// The slave is receiver
				}
			}
			// Clear SR1Register and SR2Register variables to prepare for next IT
			SR1Register = 0u ;
			SR2Register = 0u ;
		}

		// If STOPF = 1: EV4 (Slave has detected a STOP condition on the bus
		if (( SR1Register & I2C_SR1_STOPF) == I2C_SR1_STOPF)
		{
if (DBG_I2C != 0) swoSend16 ('4/', 0) ;
			__IO uint32_t temp ;
			// Read SR1 then write CR1 to clear STOPF. Have nothing to write to CR1...
			temp = pI2c->CR1 ;
			SR1Register = pI2c->SR1 ;
			pI2c->CR1 = temp ;
			SR1Register = 0u ;
			i2cEnd (pI2cIo, (pI2cIo->status != I2C_EBUSY) ? pI2cIo->status : I2C_EOK, I2C_STATE_IDLE) ;
		}

		// If TXE = 1: EV3
		if ((SR1Register & I2C_SR1_TXE) == I2C_SR1_TXE)
		{
if (DBG_I2C != 0) swoSend16 ('7/', 0) ;
			// Sometime the TXE IT is called twice...
			// The second time after all IT were disabled and the state switched to I2C_STATE_TXWAIT
			if (getFlagState (pI2cIo) == I2C_STATE_TX)
			{
				// Write data in I2C data register
				if (pI2cIo->idx < pI2cIo->size)
				{
					pI2c->DR = pI2cIo->pBuffer [pI2cIo->idx++] ;
				}
				else
				{
					// Master requests more data than slave buffer size
					// This may be due to the dual buffer of this I2C!
					// Send 0 data until master NACK or STOP
					pI2c->DR = 0u ;
					pI2cIo->idx++ ;
				}
				if ((pI2cIo->size == pI2cIo->idx)  &&  (getFlagLast (pI2cIo) != I2C_FLAG_LAST))
				{
					// End of block xfer (not the last one)
					i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_TXWAIT) ;	// The whole block has been sent
				}
			}
			SR1Register = 0u ;
		}

		// If RXNE = 1: EV2
		if ((SR1Register & I2C_SR1_RXNE) == I2C_SR1_RXNE)
		{
			if (getFlagLast (pI2cIo) == I2C_FLAG_LAST)
			{
if (DBG_I2C != 0) swoSend16 ('5/', 0) ;
				// Last block
				if (pI2cIo->idx == pI2cIo->size-2)
				{
					// To receive the last byte disable byte acknowledge
					pI2c->CR1  &= ~I2C_CR1_ACK ;
if (DBG_I2C != 0) swoSend16 ('6/', 0) ;
				}
				// Read data from I2C data register
				if (pI2cIo->idx < pI2cIo->size)
				{
					pI2cIo->pBuffer [pI2cIo->idx++] = (uint8_t) pI2c->DR ;
				}
				else
				{
					// Master sends more data than slave buffer size
					// Should never go here if the master takes into account the NACK
					pI2cIo->status = I2C_ESIZE ;
					__IO uint32_t temp = pI2c->DR ;	// Read and trash
					(void) temp ;
				}
			}
			else
			{
if (DBG_I2C != 0) swoSend16 ('8/', 0) ;
				// Not last block
				pI2cIo->pBuffer [pI2cIo->idx++] = (uint8_t) pI2c->DR ;
				if (pI2cIo->idx == pI2cIo->size)
				{
if (DBG_I2C != 0) swoSend16 ('9/', 0) ;
					// End of block
					i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_RXWAIT) ;
				}
			}
			SR1Register = 0u ;
		}
	}	// End slave mode

	else	//--------------------------------------------------------------------

	{
		// Master mode
		// If SB = 1, I2C master sent a START on the bus: EV5
		if ((SR1Register & I2C_SR1_SB) == I2C_SR1_SB)
		{
if (DBG_I2C != 0) swoSend16 ('1-', 0) ;
			// Send the slave address for transmission or reception
			// according to the configured value in the master read/write/Xfer routine
			pI2c->DR = pI2cIo->slaveAddress ;
			SR1Register = 0u ;
		}

		// If ADDR = 1, EV6
		if ((SR1Register & I2C_SR1_ADDR) == I2C_SR1_ADDR)
		{
			if (getFlagDir (pI2cIo) == I2C_FLAG_TX)
			{
if (DBG_I2C != 0) swoSend16 ('2-', 0) ;
				// Master TX
				// Clear ADDR flag
				SR1Register = pI2c->SR1 ;
				SR2Register = pI2c->SR2 ;

				if (getFlagAuto (pI2cIo) == I2C_FLAG_AUTO)
				{
					// Switch to data transmission: Enable BUF IT
					pI2c->CR2 |= I2C_CR2_ITBUFEN ;
					setFlagState (pI2cIo, I2C_STATE_TXWAIT) ;
				}
				else
				{
					// Slave addressing phase OK, ready for data transmission
					i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_TXWAIT) ;
				}
			}
			else
			{
				// Master RX
				if (getFlagState (pI2cIo) == I2C_STATE_ADDR  &&  getFlagAuto (pI2cIo) == I2C_FLAG_SEQ)
				{
if (DBG_I2C != 0) swoSend16 ('3-', 0) ;
					// Slave addressing phase OK, ready for data receive
					// Quit without ADDR clearing
					// Return to i2cMasterStart()
					i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_RXWAIT) ;
				}
				else
				{
if (DBG_I2C != 0) swoSend16 ('4-', 0) ;
					if (getFlagAuto (pI2cIo) == I2C_FLAG_AUTO)
					{
						// Switch to data transmission: Enable BUF IT
						pI2c->CR2 |= I2C_CR2_ITBUFEN ;
					}
					// For the 1st data RX block after ADDR
					// EV6_1: used for single byte reception.
					// The ACK disable and the STOP programming should be done just after ADDR is cleared.
					// Come from i2cMasterXfer()
					if (pI2cIo->size == 1u  &&  (getFlagLast (pI2cIo) == I2C_FLAG_LAST))
					{
if (DBG_I2C != 0) swoSend16 ('5-', 0) ;
						__disable_irq();
						// Clear ADDR flag
						SR1Register = pI2c->SR1 ;
						SR2Register = pI2c->SR2 ;
						// Clear ACK
						pI2c->CR1 &= ~I2C_CR1_ACK ;
						if (getFlagAuto (pI2cIo) == I2C_FLAG_AUTO)
						{
							pI2c->CR1 |= I2C_CR1_STOP ;		// Program the STOP
						}
						__enable_irq();
					}
					else
					{
						// Other sizes : clear ADDR flag
if (DBG_I2C != 0) swoSend16 ('6-', 0) ;
						SR1Register = pI2c->SR1 ;
						SR2Register = pI2c->SR2 ;
					}
					setFlagState (pI2cIo, I2C_STATE_RX) ;
				}
			}

			SR1Register = 0u ;
			SR2Register = 0u ;
		}

		// Master transmits the remaining data until the last one.
		if (pI2cIo->size != pI2cIo->idx  &&  (SR1Register & I2C_SR1_TXE) == I2C_SR1_TXE)
		{
if (DBG_I2C != 0) swoSend16 ('7-', 0) ;
			// TxE flag and there is still data to write
			// If only 1 data remains to write, disable the BUF IT in order
			// to not have again a TxE interrupt.
			if (pI2cIo->size == pI2cIo->idx + 1u)
			{
				// Last byte to send, Disable the BUF IT and wait for BTF
				pI2c->CR2 &= ~I2C_CR2_ITBUFEN ;
			}
			// Write the data in DR register
			pI2c->DR = pI2cIo->pBuffer [pI2cIo->idx++] ;
			SR1Register = 0u ;
		}

		// If BTF and TXE are set (EV8_2) then end of xfer
		if ((SR1Register & (I2C_SR1_TXE | I2C_SR1_BTF)) == (I2C_SR1_TXE | I2C_SR1_BTF))
		{
if (DBG_I2C != 0) swoSend16 ('8-', 0) ;
			if (getFlagAuto (pI2cIo) == I2C_FLAG_AUTO)
			{
				pI2c->CR1 |= I2C_CR1_STOP ;		// Program the STOP
			}
			i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_TXWAIT) ;	// End of block
			SR1Register = 0u ;
		}

		// If RXNE is set
		if ((SR1Register & I2C_SR1_RXNE) == I2C_SR1_RXNE)
		{
			// Read the data register
			if ((pI2cIo->idx == pI2cIo->size - 2u)  &&  (getFlagLast (pI2cIo) == I2C_FLAG_LAST))
			{
if (DBG_I2C != 0) swoSend16 ('9-', 0) ;
				// Non interruptible sequence: read DR + Clear ACK + Set STOP
				__disable_irq();
				pI2cIo->pBuffer [pI2cIo->idx++] = (uint8_t) pI2c->DR ;
				// It remains only one byte to receive, disable ACK and program the STOP (EV7_1)
				pI2c->CR1 &= ~I2C_CR1_ACK ;		// Clear ACK
				if (getFlagAuto (pI2cIo) == I2C_FLAG_AUTO)
				{
if (DBG_I2C != 0) swoSend16 ('a-', 0) ;
					pI2c->CR1 |= I2C_CR1_STOP ;		// Program the STOP
				}
				__enable_irq();
			}
			else
			{
if (DBG_I2C != 0) swoSend16 ('b-', 0) ;
				pI2cIo->pBuffer [pI2cIo->idx++] = (uint8_t) pI2c->DR ;
			}

			// Check end of block
			if (pI2cIo->idx == pI2cIo->size)
			{
if (DBG_I2C != 0) swoSend16 ('c-', 0) ;
				i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_RXWAIT) ;	// End of block
			}
			SR1Register = 0u ;
		}
	}	 // End master mode

	aaIntExit () ;
}

//----------------------------------------------------------------------

#if (WITH_I2C_1 == 1)
void I2C1_EV_IRQHandler (void) ;
void I2C1_EV_IRQHandler (void)
{
	I2Cx_EV_IRQHandler (I2C1, & i2cIo1) ;
}
#endif	// WITH_I2C_1

#if (WITH_I2C_2 == 1)
void I2C2_EV_IRQHandler (void) ;
void I2C2_EV_IRQHandler (void)
{
	I2Cx_EV_IRQHandler (I2C2, & i2cIo2) ;
}
#endif	// WITH_I2C_2

#if (WITH_I2C_3 == 1)
void I2C3_EV_IRQHandler (void) ;
void I2C3_EV_IRQHandler (void)
{
	I2Cx_EV_IRQHandler (I2C3, & i2cIo3) ;
}
#endif	// WITH_I2C_3

#if (WITH_I2C_4 == 1  &&  defined I2C4)
void I2C4_EV_IRQHandler (void) ;
void I2C4_EV_IRQHandler (void)
{
	I2Cx_EV_IRQHandler (I2C4, & i2cIo4) ;
}
#endif	// WITH_I2C_4

//----------------------------------------------------------------------

void I2Cx_ER_IRQHandler (I2C_TypeDef * pI2c, i2cIo_t * pI2cIo) ;

void I2Cx_ER_IRQHandler (I2C_TypeDef * pI2c, i2cIo_t * pI2cIo)
{
	__IO uint32_t	SR1Register ;
	uint32_t		status ;

	aaIntEnter () ;

	// Read the I2C1 status register
	SR1Register = pI2c->SR1 ;

	// If AF = 1
	if ((SR1Register & I2C_SR1_AF) == I2C_SR1_AF)
	{
		pI2c->SR1 &= ~I2C_SR1_AF ;	// Clear AF
		if (getFlagMode (pI2cIo) == I2C_FLAG_SLAVE)	//--------------------------------
		{
if (DBG_I2C != 0) swoSend16 ('a/', 0) ;
			// I2C is slave (MSL flag = 0), and slave is TX
			if ((SR1Register & I2C_SR1_TXE) == 0u)
			{
				// One byte in DR was not sent
				if (pI2cIo->idx > 0u)
				{
					pI2cIo->idx-- ;
				}
			}
			if (pI2cIo->idx >= pI2cIo->size)
			{
				// Size bytes was transfered, so the transfer size is OK
				status = I2C_EOK ;
			}
			else
			{
				// The master stopped receive before end of buffer
				status = I2C_EAF ;
			}
			i2cEnd (pI2cIo, status, I2C_STATE_IDLE) ;
		}
		else		//-------------------------------------------------------------------
		{
			// I2C is master TX. The slave doesn't acknowledge addressing
			if (getFlagState (pI2cIo) == I2C_STATE_ADDR)
			{
				// AF in addressing phase
if (DBG_I2C != 0) swoSend16 ('x-', 0) ;
				i2cEnd (pI2cIo, I2C_EAF, I2C_STATE_IDLE) ;
			}
			else
			{
				// AF in TX data phase
if (DBG_I2C != 0) swoSend16 ('y-', 0) ;
				if ((SR1Register & I2C_SR1_TXE) == 0u)
				{
					// One byte in DR was not sent
					if (pI2cIo->idx > 0u)
					{
						pI2cIo->idx-- ;
					}
				}
				// End of master transfer, then must send STOP or ReSTART
				i2cEnd (pI2cIo, (pI2cIo->idx == pI2cIo->size) ? I2C_EOK : I2C_EAF, I2C_STATE_IDLE) ;
				if (getFlagAuto (pI2cIo) == I2C_FLAG_AUTO)
				{
					// Program the STOP
if (DBG_I2C != 0) swoSend16 ('z-', 0) ;
					pI2c->CR1 |= I2C_CR1_STOP ;
				}
			}
		}		//-------------------------------------------------------------------

		SR1Register = 0u ;
	}

	// If ARLO = 1
	if ((SR1Register & I2C_SR1_ARLO) == I2C_SR1_ARLO)
	{
		pI2c->SR1 &= ~I2C_SR1_ARLO ;
		SR1Register = 0 ;
		pI2cIo->status = I2C_EARLO ;
	}
	// If BERR = 1
	if ((SR1Register & I2C_SR1_BERR) == I2C_SR1_BERR)
	{
		pI2c->SR1 &= ~I2C_SR1_BERR ;
		SR1Register = 0 ;
		pI2cIo->status = I2C_EBERR ;
	}

	// If OVR = 1. Only for slave with clock stretching disabled.
	if ((SR1Register & I2C_SR1_OVR) == I2C_SR1_OVR)
	{
		pI2c->SR1 &= ~I2C_SR1_OVR ;
		SR1Register = 0 ;
		pI2cIo->status = I2C_EOVR ;
	}

	aaIntExit () ;
}

//----------------------------------------------------------------------

#if (WITH_I2C_1 == 1)
void I2C1_ER_IRQHandler(void) ;
void I2C1_ER_IRQHandler(void)
{
	I2Cx_ER_IRQHandler (I2C1, & i2cIo1) ;
}
#endif	// WITH_I2C_1

#if (WITH_I2C_2 == 1)
void I2C2_ER_IRQHandler (void) ;
void I2C2_ER_IRQHandler (void)
{
	I2Cx_ER_IRQHandler (I2C2, & i2cIo2) ;
}
#endif	// WITH_I2C_2

#if (WITH_I2C_3 == 1)
void I2C3_ER_IRQHandler (void) ;
void I2C3_ER_IRQHandler (void)
{
	I2Cx_ER_IRQHandler (I2C3, & i2cIo3) ;
}
#endif	// WITH_I2C_3

#if (WITH_I2C_4 == 1  &&  defined I2C4)
void I2C4_ER_IRQHandler (void) ;
void I2C4_ER_IRQHandler (void)
{
	I2Cx_ER_IRQHandler (I2C4, & i2cIo4) ;
}
#endif	// WITH_I2C_4

//----------------------------------------------------------------------
#if (DBG_I2C != 0)
	#pragma GCC diagnostic pop
#endif
