/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	gpiobasic.c	Some routine to manage basic GPIO and RCC features

	When		Who	What
	17/03/18	ac	Creation for STM32F4xx
	27/06/20	ac	Bug in rccGetTimerClockFrequency(): bad prescaler comparison value.
	04/03/21	ac	RCC_GetXxxClockFreq() functions are now static inside stm32l4xx_ll_rcc.c. Sigh...
					So create public rccGetXxxClockFreq()
	01/02/22	ac	rccbasic split from gpiobasic

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2023 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#include	"aa.h"

#include	"stm32f4xx_ll_bus.h"
#include	"stm32f4xx_ll_rcc.h"

//----------------------------------------------------------------------

#define	OFFSET_MASK		0x0000FFFFu

//----------------------------------------------------------------------
//	Some utility functions to get clocks frequencies

uint32_t	rccGetSystemClockFreq	(void)
{
	// SYSTEM clock frequency
	return SystemCoreClock ;
}

uint32_t	rccGetHCLKClockFreq		(void)
{
	// HCLK clock frequency
	return __LL_RCC_CALC_HCLK_FREQ (SystemCoreClock, LL_RCC_GetAHBPrescaler());
}

uint32_t	rccGetPCLK1ClockFreq	(void)
{
	// PCLK1 clock frequency
	return __LL_RCC_CALC_PCLK1_FREQ	(rccGetHCLKClockFreq (), LL_RCC_GetAPB1Prescaler());
}

#if defined __LL_RCC_CALC_PCLK2_FREQ
uint32_t	rccGetPCLK2ClockFreq	(void)
{
	// PCLK2 clock frequency
	return __LL_RCC_CALC_PCLK2_FREQ (rccGetHCLKClockFreq (), LL_RCC_GetAPB2Prescaler());
}
#endif

//----------------------------------------------------------------------
// Peripheral clock enable
// Example:
// rccEnableClock (gpioPortFromName (pPinDesc->port), 1u << (pPinDesc->port - AA_GPIO_PORTA)) ;
// rccEnableClock (I2C3, RCC_APB1ENR_I2C3EN) ;

void	rccEnableClock (void * deviceAddress, uint32_t enableBitMask)
{
	switch ((uintptr_t) deviceAddress & ~OFFSET_MASK)
	{
		case APB1PERIPH_BASE:
			LL_APB1_GRP1_EnableClock (enableBitMask) ;	// e.g. LL_APB1_GRP1_PERIPH_I2C2
			break ;

		case APB2PERIPH_BASE:
			LL_APB2_GRP1_EnableClock (enableBitMask) ;	// e.g. LL_APB2_GRP1_PERIPH_TIM9
			break ;

#if defined(AHB1PERIPH_BASE)
		case AHB1PERIPH_BASE:
			LL_AHB1_GRP1_EnableClock (enableBitMask) ;
			break ;
#endif

#if defined(RCC_AHB2_SUPPORT)
		case AHB2PERIPH_BASE:
			LL_AHB2_GRP1_EnableClock (enableBitMask) ;
			break ;
#endif

		default:
			AA_ASSERT (0) ;
			break ;
	}
}

//----------------------------------------------------------------------
void	rccDisableClock (void * deviceAddress, uint32_t enableBitMask)
{
	switch ((uintptr_t) deviceAddress & ~OFFSET_MASK)
	{
		case APB1PERIPH_BASE:
			LL_APB1_GRP1_DisableClock (enableBitMask) ;	// LL_APB1_GRP1_PERIPH_I2C2
			break ;

		case APB2PERIPH_BASE:
			LL_APB2_GRP1_DisableClock (enableBitMask) ;	// LL_APB1_GRP1_PERIPH_I2C2
			break ;

#if defined(AHB1PERIPH_BASE)
		case AHB1PERIPH_BASE:
			LL_AHB1_GRP1_DisableClock (enableBitMask) ;
			break ;
#endif

#if defined(RCC_AHB2_SUPPORT)
		case AHB2PERIPH_BASE:
			LL_AHB2_GRP1_DisableClock (enableBitMask) ;
			break ;
#endif

		default:
			AA_ASSERT (0) ;
			break ;
	}
}

//----------------------------------------------------------------------
//	Returns clock frequency of a device, except for timers

uint32_t	rccGetClockFrequency (void * deviceAddress)
{
	LL_RCC_ClocksTypeDef	clocks ;
	uint32_t				devClock ;

	LL_RCC_GetSystemClocksFreq (& clocks) ;

	switch ((uintptr_t)  deviceAddress & ~OFFSET_MASK)
	{
		case APB1PERIPH_BASE:
			devClock = clocks.PCLK1_Frequency ;
			break ;

		case APB2PERIPH_BASE:
			devClock = clocks.PCLK2_Frequency ;
			break ;

#if defined(AHB1PERIPH_BASE)
		case AHB1PERIPH_BASE:
			devClock = clocks.HCLK_Frequency ;
			break ;
#endif

#if defined(RCC_AHB2_SUPPORT)
		case AHB2PERIPH_BASE:
			devClock = clocks.HCLK_Frequency ;
			break ;
#endif

		default:
			AA_ASSERT (0) ;
			devClock = 0u ;
			break ;
	}
	return devClock ;
}

//----------------------------------------------------------------------
//	Timer clock depends on APB bus prescaler value
//	Timer clock may be APBx peripherals frequency x2

uint32_t	rccGetTimerClockFrequency (void * deviceAddress)
{
	LL_RCC_ClocksTypeDef	clocks ;
	uint32_t				prescaler ;
	uint32_t				timClock ;

	LL_RCC_GetSystemClocksFreq (& clocks) ;

	switch ((uintptr_t)  deviceAddress & ~OFFSET_MASK)
	{
		case APB1PERIPH_BASE:
			timClock = clocks.PCLK1_Frequency ;
			prescaler = LL_RCC_GetAPB1Prescaler () ;
			if (prescaler >= RCC_CFGR_PPRE1_DIV2)
			{
				timClock *= 2u ;
			}
			break ;

		case APB2PERIPH_BASE:
			timClock = clocks.PCLK2_Frequency ;
			prescaler = LL_RCC_GetAPB2Prescaler () ;
			if (prescaler >= RCC_CFGR_PPRE2_DIV2)
			{
				timClock *= 2u ;
			}
			break ;

		default:
			AA_ASSERT (0) ;
			timClock = 0u ;
			break ;
	}
	return timClock ;
}

//----------------------------------------------------------------------

void	rccResetPeriph (void * deviceAddress, uint32_t enableBitMask)
{
	switch ((uintptr_t) deviceAddress & ~OFFSET_MASK)
	{
		case APB1PERIPH_BASE:
			LL_APB1_GRP1_ForceReset   (enableBitMask) ;
			LL_APB1_GRP1_ReleaseReset (enableBitMask) ;
			break ;

		case APB2PERIPH_BASE:
			LL_APB2_GRP1_ForceReset   (enableBitMask) ;
			LL_APB2_GRP1_ReleaseReset (enableBitMask) ;
			break ;

#if defined(AHB1PERIPH_BASE)
		case AHB1PERIPH_BASE:
			LL_AHB1_GRP1_ForceReset   (enableBitMask) ;
			LL_AHB1_GRP1_ReleaseReset (enableBitMask) ;
			break ;
#endif

#if defined(RCC_AHB2_SUPPORT)
		case AHB2PERIPH_BASE:
			LL_AHB2_GRP1_ForceReset   (enableBitMask) ;
			LL_AHB2_GRP1_ReleaseReset (enableBitMask) ;
			break ;
#endif

		default:
			AA_ASSERT (0) ;
			break ;
	}
}

//----------------------------------------------------------------------
