/*
----------------------------------------------------------------------

	Alain Chebrou

	RTC.c		Manage the RTC of some STM32
				F0 F2 F3 F4 F7 L0 L1 H72/H73/H74/H75
				L4  except L41/L42
				L4+ except L4P5/Q5
				AN3371, AN4759

	When		Who	What
	26/06/21	ac	Creation

	RTC use LSE, hour 24, calendar, alarm, backup registers access. Lacks time stamp, wake up.

----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aakernel.h"		// For aaIntenter() / aaIntExit()
#include	"aalogmes.h"
#include	"aaprintf.h"

#include	"rtcbasic.h"
#include	"extibasic.h"

#if defined STM32F4
	#include	"stm32f4xx_ll_rcc.h"		// Enable clocks
	#include	"stm32f4xx_ll_rtc.h"
	#include	"stm32f4xx_ll_pwr.h"		// LL_PWR_EnableBkUpAccess
#endif

#if defined STM32L4
	#include	"stm32l4xx_ll_rcc.h"		// Enable clocks
	#include	"stm32l4xx_ll_rtc.h"
	#include	"stm32l4xx_ll_bus.h"		// Enable PWR
	#include	"stm32l4xx_ll_pwr.h"		// LL_PWR_EnableBkUpAccess
#endif

#if defined STM32H7
	#include	"stm32h7xx_ll_rcc.h"		// Enable clocks
	#include	"stm32h7xx_ll_rtc.h"
	#include	"stm32h7xx_ll_pwr.h"		// LL_PWR_EnableBkUpAccess
#endif

// The RTC alarm EXTI line is not the same on all families...
#if (defined STM32F4  ||  defined STM32H7)
	#define	RTC_ALARM_EXTI		17
#endif
#if (defined STM32L4)
	#define	RTC_ALARM_EXTI		18
#endif

//--------------------------------------------------------------------------------
// For LSE clock source: ck_spre = ck_lse / (ASYNC prediv + 1) / (SYNC prediv + 1) = 1 Hz
#define RTC_ASYNCH_PREDIV	((uint32_t)0x7F)
#define RTC_SYNCH_PREDIV	((uint32_t)0x00FF)

static	void	(* alarmACallback) (void) ;
static	void	(* alarmBCallback) (void) ;

// For internal use. Most of the time applications doesn't need these functions
static	uint32_t	rtcWaitForSynchro		(void) ;
static	void		rtcEnterInitMode		(void) ;
static	uint32_t	rtcExitInitMode			(void) ;

//--------------------------------------------------------------------------------

BSP_ATTR_ALLWAYS_INLINE	void	rtcEnableAccess (void)
{
	LL_PWR_EnableBkUpAccess () ;
	LL_RTC_DisableWriteProtection (RTC) ;
}

BSP_ATTR_ALLWAYS_INLINE	void	rtcDisableAccess (void)
{
	LL_PWR_DisableBkUpAccess () ;
	LL_RTC_EnableWriteProtection (RTC) ;
}

//--------------------------------------------------------------------------------
//	Reset all the backup domain

void	rtcResetBackup (void)
{
	LL_PWR_EnableBkUpAccess () ;
	LL_RCC_ForceBackupDomainReset () ;
	LL_RCC_ReleaseBackupDomainReset () ;
#if (! defined STM32L4)
	LL_PWR_DisableBkUpRegulator () ;
#endif
	LL_PWR_DisableBkUpAccess () ;
}

//--------------------------------------------------------------------------------
//	Return: 1 if RTC is enabled and the calendar is configured

uint32_t	rtcCheckConfig ()
{
	if ((RTC->ISR & RTC_ISR_INITS) != 0)
	{
		return 1 ;
	}
	return 0 ;
}

//--------------------------------------------------------------------------------
//	Wait until the RTC Time and Date registers (RTC_TR and RTC_DR) are synchronized with RTC APB clock

static	uint32_t rtcWaitForSynchro (void)
{
  // Clear RSF flag
  LL_RTC_ClearFlag_RS (RTC) ;

  // Wait the registers to be synchronized
  while (LL_RTC_IsActiveFlag_RS (RTC) != 1)
  {
  }
  return RTC_ENONE ;
}

//--------------------------------------------------------------------------------
//	Enter RTC initialization mode

static	void rtcEnterInitMode (void)
{
	// Set Initialization mode
	LL_RTC_EnableInitMode (RTC) ;

	// Check if the Initialization mode is set
	while (LL_RTC_IsActiveFlag_INIT (RTC) != 1)
	{
	}
}

//--------------------------------------------------------------------------------
//	Exit RTC Initialization mode

static	uint32_t rtcExitInitMode (void)
{
  LL_RTC_DisableInitMode (RTC) ;

  // Wait for synchro
  // Note: Needed only if Shadow registers is enabled
  //       LL_RTC_IsShadowRegBypassEnabled function can be used
  return rtcWaitForSynchro () ;
}

//--------------------------------------------------------------------------------
//	Write 1 backup register. Need Backup Access
//	num is: 0..RTC_BKP_NUMBER

void	rtcWriteBackupRegister (uint32_t num, uint32_t value)
{
	rtcEnableAccess () ;
	AA_ASSERT (num < RTC_BKP_NUMBER) ;
	if (num < RTC_BKP_NUMBER)
	{
		(& (RTC->BKP0R)) [num] = value ;
	}
	rtcDisableAccess () ;
}

//--------------------------------------------------------------------------------
//	Configure LSE clock for RTC
//	Must be called once at application startup
//	Return: 1 if the RTC need to be configured (Backup domain was reseted)

uint32_t rtcConfigureClocks (void)
{
#if (defined STM32F4)
	__IO uint32_t	tmpreg ;

	// Enable power clock
	RCC->APB1ENR |= RCC_APB1ENR_PWREN ;
	tmpreg = RCC->APB1ENR ;	// Delay after an RCC peripheral clock enabling
	(void) tmpreg ;
#endif

#if (defined STM32L4)
	LL_APB1_GRP1_EnableClock(LL_APB1_GRP1_PERIPH_PWR);
#endif

	// To allow access to LSEON, LSEBYP, RTCSEL and RTCEN bits in RCC->BDCR
	LL_PWR_EnableBkUpAccess () ;

	// Configure LSE as RTC clock source
	// Enable LSE only if it is disabled
	if (LL_RCC_LSE_IsReady() == 0)
	{
		LL_RCC_ForceBackupDomainReset () ;
		LL_RCC_ReleaseBackupDomainReset () ;
#if (! defined STM32L4)
		LL_PWR_EnableBkUpRegulator () ;
#endif
//		LL_RCC_LSE_EnableHighDriveMode () ;
		LL_RCC_LSE_Enable () ;
		while (LL_RCC_LSE_IsReady() != 1)
		{
		}
		LL_RCC_SetRTCClockSource (LL_RCC_RTC_CLKSOURCE_LSE) ;
		LL_RCC_EnableRTC () ;
		LL_PWR_DisableBkUpAccess () ;
		return 1 ;				// RTC must be configured
	}

	LL_PWR_DisableBkUpAccess () ;
	return 0 ;	// RTC is already configured
}

//--------------------------------------------------------------------------------
//	Configure RTC for calendar. Doesn't configure RTC output.
//	If asyncDiv and syncDiv are 0, then use the default divisors for LSE

void rtcConfigure (uint32_t hourFormat, uint32_t asyncDiv, uint32_t syncDiv)
{
	rtcEnableAccess () ;

	// Enter initialization mode
	rtcEnterInitMode () ;

	// Configure RTC

	// Configure RTC prescaler and RTC data registers
	// Set Hour Format
	LL_RTC_SetHourFormat (RTC, hourFormat) ;

	if ((asyncDiv == 0u)  &&  (syncDiv == 0u))
	{
		// Configure for LSE standard frequency
		// 1st Set Synch Prediv
		LL_RTC_SetSynchPrescaler (RTC, RTC_SYNCH_PREDIV) ;

		// 2nd Set Asynch Prediv
		LL_RTC_SetAsynchPrescaler (RTC, RTC_ASYNCH_PREDIV) ;
	}
	else
	{
		// Configure with specified divisors
		// 1st Set Synch Prediv
		LL_RTC_SetSynchPrescaler (RTC, syncDiv) ;

		// 2nd Set Asynch Prediv
		LL_RTC_SetAsynchPrescaler (RTC, asyncDiv) ;
	}
	// Configure RTC outPut
	// Reset value is LL_RTC_ALARMOUT_DISABLE
//	LL_RTC_SetAlarmOutEvent (RTC, LL_RTC_ALARMOUT_DISABLE) ;

	// Set OutPutPolarity
	// Reset value is LL_RTC_OUTPUTPOLARITY_PIN_HIGH
//	LL_RTC_SetOutputPolarity (RTC, LL_RTC_OUTPUTPOLARITY_PIN_HIGH) ;

	// Set OutPutType
	// Reset value is LL_RTC_ALARM_OUTPUTTYPE_OPENDRAIN
//	LL_RTC_SetAlarmOutputType (RTC, LL_RTC_ALARM_OUTPUTTYPE_OPENDRAIN) ;

	// Exit initialization mode
	rtcExitInitMode () ;

	rtcDisableAccess () ;
}

//--------------------------------------------------------------------------------
//	Configure the current time and date (binary)
//	Time word	0x00HHMMSS
//	Date word	0xDDMMYYYY	=> with full year
//	Week day	1 to 7, Monday to Sunday

void rtcSetCalendar (uint32_t weekDay, uint32_t time, uint32_t date)
{
	uint32_t	reg, year ;

	AA_ASSERT (weekDay > 0  &&  weekDay < 8) ;

	rtcEnableAccess () ;

	// Enter in initialization mode
	rtcEnterInitMode () ;

	// Configure the Date
	// __LL_RTC_CONVERT_BIN2BCD helper macro is used to convert values from bin to BCD
	year = RTC_GETYEAR  (date) ;
	if (year > 99)
	{
		year %= 100 ;
	}
	year = __LL_RTC_CONVERT_BIN2BCD (year) ;
	reg =	(weekDay << RTC_DR_WDU_Pos)   |
			(year << RTC_DR_YU_Pos)	      |
			(__LL_RTC_CONVERT_BIN2BCD (RTC_GETMONTH (date)) << RTC_DR_MU_Pos)	|
			(__LL_RTC_CONVERT_BIN2BCD (RTC_GETDAY   (date)) << RTC_DR_DU_Pos) ;
	RTC->DR = reg ;

	reg =	(__LL_RTC_CONVERT_BIN2BCD (RTC_GETHOUR  (time)) << RTC_TR_HU_Pos)	|
			(__LL_RTC_CONVERT_BIN2BCD (RTC_GETMIN   (time)) << RTC_TR_MNU_Pos)	|
			(__LL_RTC_CONVERT_BIN2BCD (RTC_GETSEC   (time)) << RTC_TR_SU_Pos) ;
	RTC->TR  = reg ;
	RTC->SSR = 0 ;

	// Exit of initialization mode
	rtcExitInitMode () ;

	rtcDisableAccess () ;
}

//--------------------------------------------------------------------------------
//	Get the current time and date (binary)

//	Time word	0x00HHMMSS
//	Date word	0xDDMMYYYY	=> with full year
//	Week day	1 to 7, Monday to Sunday

void rtcGetCalendar (uint32_t * pWeekDay, uint32_t * pTime, uint32_t * pDate)
{
	uint32_t	temp, reg ;
	// Note: need to convert from BCD to decimal value using __LL_RTC_CONVERT_BCD2BIN helper macro

	reg = RTC->TR ;
	if (pTime != NULL)
	{
		temp  = __LL_RTC_CONVERT_BCD2BIN ((reg & (RTC_TR_HT  | RTC_TR_HU))  >> RTC_TR_HU_Pos)  << 16 ;	// Hour
		temp |= __LL_RTC_CONVERT_BCD2BIN ((reg & (RTC_TR_MNT | RTC_TR_MNU)) >> RTC_TR_MNU_Pos) << 8 ;	// Min
		temp |= __LL_RTC_CONVERT_BCD2BIN ((reg & (RTC_TR_ST  | RTC_TR_SU))  >> RTC_TR_SU_Pos) ;			// Sec
		* pTime = temp ;
	}

	reg = RTC->DR ;
	if (pDate != NULL)
	{
		temp  = __LL_RTC_CONVERT_BCD2BIN ((reg &  (RTC_DR_DT | RTC_DR_DU)) >> RTC_DR_DU_Pos) << 24;		// Day
		temp |= __LL_RTC_CONVERT_BCD2BIN ((reg &  (RTC_DR_MT | RTC_DR_MU)) >> RTC_DR_MU_Pos) << 16;		// Month
		temp |= __LL_RTC_CONVERT_BCD2BIN ((reg &  (RTC_DR_YT | RTC_DR_YU)) >> RTC_DR_YU_Pos) + 2000 ;	// Year
		* pDate = temp ;
	}

	if (pWeekDay != NULL)
	{
		* pWeekDay = (reg & RTC_DR_WDU) >> RTC_DR_WDU_Pos ;
	}
/*
	//	Test only: reference value
	aaPrintf ("RTC %02u:%02u:%02u ",
			__LL_RTC_CONVERT_BCD2BIN (LL_RTC_TIME_GetHour   (RTC)),
			__LL_RTC_CONVERT_BCD2BIN (LL_RTC_TIME_GetMinute (RTC)),
			__LL_RTC_CONVERT_BCD2BIN (LL_RTC_TIME_GetSecond (RTC))) ;

	aaPrintf ("%u %02u-%02u-%02u\n",
			LL_RTC_DATE_GetWeekDay (RTC),
			__LL_RTC_CONVERT_BCD2BIN (LL_RTC_DATE_GetDay   (RTC)),
			__LL_RTC_CONVERT_BCD2BIN (LL_RTC_DATE_GetMonth (RTC)),
			2000 + __LL_RTC_CONVERT_BCD2BIN (LL_RTC_DATE_GetYear (RTC)));
*/
}

//--------------------------------------------------------------------------------

void rtcPrintTime (uint32_t time)
{
	aaPrintf ("%02u:%02u:%02u", RTC_GETHOUR(time), RTC_GETMIN(time), RTC_GETSEC(time)) ;
}

void rtcPrintDate (uint32_t date)
{
	aaPrintf ("%02u-%02u-%04u", RTC_GETDAY(date), RTC_GETMONTH(date), RTC_GETYEAR(date)) ;
}

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//	Build time with: RTC_TOTIME (hh, mm, ss)
//	Build ssr  with: RTC_TOSSR (mask, ss)		mask is 0 to 15, 0 => ignore SS
//	date is: 1-7 for week day, or 1-31 for month day

// 	To set alarm A to 12:30:00:
//		rtcSetAlarm (RTC_TOTIME (12, 30, 00), 0, 0, RTC_ALARMA | RTC_IGNORE_WDAY, BSP_IRQPRIOMIN_PLUS(1)) ;

// 	To set alarm A every 250ms second (mask=6 => ignore SS[14:6]) with PRECS=255, see AN3371
//		rtcSetAlarm (0, RTC_TOSSR (6, 255), 0, RTC_ALARMA | RTC_IGNORE_WDAY | RTC_IGNORE_HOUR | RTC_IGNORE_MIN | RTC_IGNORE_SEC, BSP_IRQPRIOMIN_PLUS(1)) ;

void	rtcAlarmStart (uint32_t time, uint32_t ssr, uint32_t date, uint32_t flags, uint32_t irqPriority)
{
	uint32_t reg = 0 ;

	rtcEnableAccess () ;

	// Get access to alarm registers
	if ((flags & RTC_ALARMA) != 0)
	{
		RTC->CR &= ~(RTC_CR_ALRAE | RTC_CR_ALRAIE) ;
		LL_RTC_ClearFlag_ALRA (RTC) ;	// Clear the Alarm interrupt pending bit
		while ((RTC->ISR & RTC_ISR_ALRAWF) == 0)
		{
		}
	}
	else
	{
		RTC->CR &= ~(RTC_CR_ALRBE | RTC_CR_ALRBIE) ;
		LL_RTC_ClearFlag_ALRB (RTC) ;	// Clear the Alarm interrupt pending bit
		while ((RTC->ISR & RTC_ISR_ALRBWF) == 0)
		{
		}
	}

	// Set time format
	reg |= flags & RTC_FORMATPM ;

	// Set time
	reg |=	((__LL_RTC_CONVERT_BIN2BCD (RTC_GETHOUR (time)) & 0x3Fu) << RTC_ALRMAR_HU_Pos)	|
			((__LL_RTC_CONVERT_BIN2BCD (RTC_GETMIN  (time)) & 0x7Fu) << RTC_ALRMAR_MNU_Pos)	|
			((__LL_RTC_CONVERT_BIN2BCD (RTC_GETSEC  (time)) & 0x7Fu) << RTC_ALRMAR_SU_Pos) ;

	// Set day/weekday
	reg |= (__LL_RTC_CONVERT_BIN2BCD (date) & 0x3Fu) << RTC_ALRMAR_DU_Pos ;
	reg |= flags & RTC_WEEKMODE ;

	// Set don't care bits
	reg |= flags & (RTC_IGNORE_SEC | RTC_IGNORE_MIN | RTC_IGNORE_HOUR | RTC_IGNORE_WDAY) ;


	// Configure RTC Alarm EXTI Interrupt
	extiInitLine (RTC_ALARM_EXTI, irqPriority, EXTI_FLAG_INTR | EXTI_FLAG_RISING) ;

	// Set Alarm registers
	if ((flags & RTC_ALARMA) != 0)
	{
		RTC->ALRMAR   = reg ;
		RTC->ALRMASSR = ssr ;			// Set sub second configuration
		LL_RTC_ClearFlag_ALRA (RTC) ;	// Clear the Alarm interrupt pending bit
		LL_RTC_EnableIT_ALRA (RTC) ;	// Enable alarm IT
		LL_RTC_ALMA_Enable (RTC) ;		// Enable alarm
 	}
	else
	{
		RTC->ALRMBR   = reg ;
		RTC->ALRMBSSR = ssr ;			// Set sub second configuration
		LL_RTC_ClearFlag_ALRB (RTC) ;	// Clear the Alarm interrupt pending bit
		LL_RTC_EnableIT_ALRB (RTC) ;	// Enable alarm IT
		LL_RTC_ALMB_Enable (RTC) ;		// Enable alarm
	}

	rtcDisableAccess () ;
}

//--------------------------------------------------------------------------------
//	flag is RTC_ALARMA or RTC_ALARMB

void	rtcAlarmStop (uint32_t flag)
{
	rtcEnableAccess () ;
	if (flag == RTC_ALARMA)
	{
		LL_RTC_ALMA_Disable (RTC) ;
	}
	else
	{
		LL_RTC_ALMB_Disable (RTC) ;
	}
	rtcDisableAccess () ;
}

//--------------------------------------------------------------------------------

void	rtcAlarmSetCallback (uint32_t flags, void (* pCB) (void))
{
	if ((flags & RTC_ALARMA) != 0u)
	{
		alarmACallback = pCB ;
	}
	else
	{
		alarmBCallback = pCB ;
	}
}

//--------------------------------------------------------------------------------
//	RTC Alarm (A and B) interrupt handler through EXTI Line 17
//	This example doesn't use the EXTI  RTC_Alarm_IRQHandler
//	(in extibasic.h WITH_EXTI_RTC_ALRM is defined as 0)
//	=> faster than chaining 2 callback

void RTC_Alarm_IRQHandler(void)
{
	aaIntEnter () ;
	rtcEnableAccess () ;

	// If the alarm A interrupt is enabled and pending
	if (LL_RTC_IsEnabledIT_ALRA (RTC) != 0  &&  LL_RTC_IsActiveFlag_ALRA (RTC) != 0)
	{
		if (alarmACallback != NULL)
		{
			(* alarmACallback) () ;
		}
		LL_RTC_ClearFlag_ALRA (RTC) ;	// Clear the Alarm interrupt pending bit
	}

	// If the alarm B interrupt is enabled and pending
	if (LL_RTC_IsEnabledIT_ALRB (RTC) != 0  &&  LL_RTC_IsActiveFlag_ALRB (RTC) != 0)
	{
		if (alarmBCallback != NULL)
		{
			(* alarmBCallback) () ;
		}
		LL_RTC_ClearFlag_ALRB (RTC) ;	// Clear the Alarm interrupt pending bit
	}

	// Clear the EXTI's Flag for RTC Alarm
#if (defined STM32F4)
	EXTI->PR = EXTI_PR_PR17_Msk ;
#endif
#if (defined STM32L4)
	EXTI->PR1 = EXTI_PR1_PIF18_Msk ;
#endif
#if (defined STM32H7)
	EXTI->PR1 = EXTI_PR1_PR17_Msk ;
#endif

	rtcDisableAccess () ;
	aaIntExit () ;
}

//--------------------------------------------------------------------------------
