/*
----------------------------------------------------------------------

	Alain Chebrou

	rtcbasic.h	Manage the RTC of STM32F4
				AN3371, AN4759

	When		Who	What
	26/06/21	ac	Creation

	RTC use LSE, hour 24, calendar and alarm

----------------------------------------------------------------------
*/

#if ! defined RTCBASIC_H_
#define RTCBASIC_H_
//--------------------------------------------------------------------------------
// time and date are always binary values packed in a word.
// Use only 24 hour format, and LSE as RTC clock source

//	Backup registers -------------------------------------------------------------
//	Write 1 backup register
//	num is: 0..RTC_BKP_NUMBER
void		rtcWriteBackupRegister	(uint32_t num, uint32_t value) ;

//	Read 1 backup register
//	num is: 0..RTC_BKP_NUMBER
BSP_ATTR_ALLWAYS_INLINE	uint32_t	rtcReadBackupRegister (uint32_t num)
{
	AA_ASSERT (num < RTC_BKP_NUMBER) ;
	return (& (RTC->BKP0R)) [num] ;
}

void	rtcResetBackup (void) ;

//	RTC calendar -------------------------------------------------------------
uint32_t	rtcCheckConfig			(void) ;
uint32_t	rtcConfigureClocks		(void) ;
void		rtcConfigure			(uint32_t hourFormat, uint32_t asyncDiv, uint32_t syncDiv) ;
void		rtcSetCalendar			(uint32_t weekDay, uint32_t time, uint32_t date) ;
void		rtcGetCalendar			(uint32_t * pWeekDay, uint32_t * pTime, uint32_t * pDate) ;

void		rtcPrintTime			(uint32_t time) ;
void		rtcPrintDate			(uint32_t date) ;

//	Macros to extract time and date components
#define	RTC_GETHOUR(time)			(((time) >> 16))
#define	RTC_GETMIN(time)			(((time) >> 8) & 0xFFu)
#define	RTC_GETSEC(time)			(((time) & 0xFFu))

#define	RTC_GETYEAR(date)			(((date) & 0xFFFFu))
#define	RTC_GETMONTH(date)			(((date) >> 16) & 0xFFu)
#define	RTC_GETDAY(date)			(((date) >> 24) & 0xFFu)

//	Macros to build configuration words
#define	RTC_TOSSR(mask, ss)			(((ss) & 0x7FFF) | (((mask) << RTC_ALRMASSR_MASKSS_Pos) & RTC_ALRMASSR_MASKSS))
#define	RTC_TOTIME(hh, mm, ss)		(((hh) << 16) | ((mm) << 8) | (ss))
#define	RTC_TODATE(dd, mm, yy)		(((dd) << 24) | ((mm) << 16) | (yy))

//	For rtcConfigure hour format parameter
#define		RTC_HOUR_24				0
#define		RTC_HOUR_AMPM			RTC_CR_FMT

//	RTC Alarm -------------------------------------------------------------
//	Flags for rtcAlarmSet
#define	RTC_ALARMA				0x00000001u
#define	RTC_ALARMB				0x00000002u

#define	RTC_FORMAT24			0x00000000u			// AM or 24 hour format
#define	RTC_FORMATPM			RTC_TR_PM			// PM time

#define	RTC_WEEKMODE			RTC_ALRMAR_WDSEL	// Date parameter is a week day (1 to 7, Monday to Sunday)

#define	RTC_IGNORE_SEC			RTC_ALRMAR_MSK1		// Seconds don't care
#define	RTC_IGNORE_MIN			RTC_ALRMAR_MSK2		// Minutes don't care
#define	RTC_IGNORE_HOUR			RTC_ALRMAR_MSK3		// Hours don't care
#define	RTC_IGNORE_WDAY			RTC_ALRMAR_MSK4		// Ignore week day or day

void	rtcAlarmSetCallback		(uint32_t flags, void (* pCB) (void));
void	rtcAlarmStart			(uint32_t time, uint32_t ssr, uint32_t date, uint32_t flags, uint32_t irqPriority) ;
void	rtcAlarmStop			(uint32_t flag);

//--------------------------------------------------------------------------------
//	Errors returned by some functions

#define RTC_ENONE				0
#define RTC_ETIMEOUT			1

//--------------------------------------------------------------------------------
#endif	// RTCBASIC_H_
