/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	spibasic.c	Master only, software or hardware SS management, no CRC
				STM32F4xx

	When		Who	What
	12/04/18	ac	Creation
	06/04/20	ac	Add 16 bits data word management
	30/05/20	ac	Error in 16 bits receive
	30/04/21	ac	Add slave mode

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2023 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aakernel.h"
#include	"gpiobasic.h"
#include	"rccbasic.h"
#include	"spibasic.h"

// Check if selected SPIs are defined for this MCU
#if (WITH_SPI_2 == 1  &&  ! defined SPI2)	||	\
	(WITH_SPI_3 == 1  &&  ! defined SPI3)	||	\
	(WITH_SPI_4 == 1  &&  ! defined SPI4)	||	\
	(WITH_SPI_5 == 1  &&  ! defined SPI5)	||	\
	(WITH_SPI_6 == 1  &&  ! defined SPI6)
	#error	"SPI does not exist in this MCU"
#endif

#include	"stm32f4xx_ll_spi.h"



#if (WITH_SPI_X == 1)
	// Include the proper SPI descriptors
	#if (defined STM32F4)
		#include "spiF4xx.h"
	#endif
#endif

#if (WITH_SPI_X == 1)
	// Count of SPI devices in spiDescArray
	STATIC	const uint32_t	spiArrayCount = sizeof (spiDescArray) / sizeof (spiDesc_t) ;
#endif

//----------------------------------------------------------------------

// Used for timing measure with an oscilloscope.
//#define dbgPin1(v)		bspOutput (BSP_OUTPUT1, v) ;	// Debug only
#define dbgPin1(v)											// No debug pulse

#define	SPI_DEF_TIMEOUT		10u			// Timeout in kernel tick

//----------------------------------------------------------------------
//----------------------------------------------------------------------
// SPI handler structure. The SPI handle is a pointer to this structure.

typedef struct
{
	const spiDesc_t	* pSpiDesc ;	// The descriptor from spiMasterConfigure()
	SPI_TypeDef		* pSpi ;		// SPI1, SPI2...
	spiStatus_t		result ;		// Result of the operation (DMA only)
	uint32_t		timeout ;		// Transfer timeout (system tick)
	aaTaskId_t		taskId ;		// The task to signal at the end of operations  (Master DMA only)

} spiIo_t ;

//	One handler structure for each physical SPI peripheral

#if (WITH_SPI_1 == 1)
	spiIo_t		spiIo1 ;
#endif	// WITH_SPI_1

#if (WITH_SPI_2 == 1)
	spiIo_t		spiIo2 ;
#endif	// WITH_SPI_2

#if (WITH_SPI_3 == 1)
	spiIo_t		spiIo3 ;
#endif	// WITH_SPI_3

#if (WITH_SPI_4 == 1  &&  defined SPI4)
	spiIo_t		spiIo4 ;
#endif	// WITH_SPI_4

#if (WITH_SPI_5 == 1  &&  defined SPI5)
	spiIo_t		spiIo5 ;
#endif	// WITH_SPI_5

#if (WITH_SPI_6 == 1  &&  defined SPI6)
	spiIo_t		spiIo6 ;
#endif	// WITH_SPI_6

//----------------------------------------------------------------------
//	Get the I/O descriptor address for an SPI peripheral

static	spiIo_t *	spiGetIoDesc (SPI_TypeDef * pSpi) ;

static	spiIo_t *	spiGetIoDesc (SPI_TypeDef * pSpi)
{
	#if (WITH_SPI_X == 0)
	{
		(void) pSpi ;	// To avoid compiler warning when none SPI is supported
	}
	#endif

	#if (WITH_SPI_1 == 1)
	{
		if (pSpi == SPI1)
		{
			return & spiIo1 ;
		}
	}
	#endif	// WITH_SPI_1

	#if (WITH_SPI_2 == 1)
	{
		if (pSpi == SPI2)
		{
			return & spiIo2 ;
		}
	}
	#endif	// WITH_SPI_2

	#if (WITH_SPI_3 == 1)
	{
		if (pSpi == SPI3)
		{
			return & spiIo3 ;
		}
	}
	#endif	// WITH_SPI_3

	#if (WITH_SPI_4 == 1  &&  defined SPI4)
	{
		if (pSpi == SPI4)
		{
			return & spiIo4 ;
		}
	}
	#endif	// WITH_SPI_4

	#if (WITH_SPI_5 == 1  &&  defined SPI5)
	{
		if (pSpi == SPI5)
		{
			return & spiIo5 ;
		}
	}
	#endif	// WITH_SPI_5

	#if (WITH_SPI_6 == 1  &&  defined SPI6)
	{
		if (pSpi == SPI6)
		{
			return & spiIo6 ;
		}
	}
	#endif	// WITH_SPI_6

	AA_ASSERT (0) ;
	return NULL ;
}

//----------------------------------------------------------------------
//	Return SPI descriptor address from an index
//	The return value is an opaque handle for the application

spiStatus_t		spiGetDesc	(uint32_t id, const spiDesc_t ** ppSpiDesc)
{
	#if (WITH_SPI_X == 1)
		uint32_t	ii ;

		for (ii = 0u ; ii < spiArrayCount ; ii++)
		{
			if (id == spiDescArray [ii].id)
			{
				* ppSpiDesc = & spiDescArray [ii] ;
				return SPI_ENONE ;
			}
		}
	#else
		(void) id ;
		(void) ppSpiDesc ;
	#endif

	return SPI_EFAIL ;
}

//----------------------------------------------------------------------
//	Set / Reset Slave Select pin on master side
//	Use NSS pin of SPI descriptor
//	SPI slave select is active low: so bSelect 1 => NSS 0

void	spiSlaveSelect (spiHandle_t hdl, uint32_t bSelect)
{
	spiIo_t		* pSpiIo = (spiIo_t *) hdl ;

	if ((pSpiIo->pSpiDesc->pinNss.port != 0u)  &&  (pSpiIo->pSpiDesc->pinNss.port != '0'))
	{
		gpioPinSet (& pSpiIo->pSpiDesc->pinNss, (bSelect & 1u) ^ 1u) ;
	}
}

//----------------------------------------------------------------------
//	Find baud rate control value from bits/s rate
//	The return value is the SPI_CR1 BR[2:0] field value

uint32_t	spiSpeedToBr (const spiDesc_t * pSpiDesc, uint32_t speed)
{
	uint32_t	br ;
	uint32_t	clock ;

	// Max allowed clock is Fpclk / 2
	clock = rccGetClockFrequency (pSpiDesc->pSpi) / 2u ;

	for (br = 0u ; br < 7u ; br++)
	{
		if (speed >= clock)
		{
			break ;
		}
		clock /= 2u ;
	}

	return br ;
}

//----------------------------------------------------------------------
//	Find bit/s rate from BR control register value

uint32_t	spiBrToSpeed (const spiDesc_t * pSpiDesc, uint32_t br)
{
	uint32_t	clock ;

	clock = rccGetClockFrequency (pSpiDesc->pSpi) ;
	return clock >> ((br & 0x7u) + 1u) ;
}

//----------------------------------------------------------------------
// Set some options for master SPI
// To use after spiMasterConfigure()

spiStatus_t		spiMasterSetOption		(spiHandle_t hdl, uint32_t option, uintptr_t value)
{
	spiIo_t			* pSpiIo = (spiIo_t *) hdl ;
//	SPI_TypeDef		* pSpi = pSpiIo->pSpi ;

	switch (option)
	{
		case SPI_OPT_TIMEOUT:
			pSpiIo->timeout = (value > 0xFFFFu) ? 0xFFFFu : (uint32_t) value ;
			break ;

		default:
			return SPI_EARG ;
	}
	return SPI_ENONE ;
}

//----------------------------------------------------------------------
// Callback to manage master SPI end of DMA transfer

static	void spiCallback (uint32_t event, uint32_t source, uintptr_t arg) ;

static	void spiCallback (uint32_t event, uint32_t source, uintptr_t arg)
{
	spiIo_t		* pSpiIo = (spiIo_t *) arg ;

dbgPin1 (1) ;
	if (event == DMA_EV_ERROR)
	{
		if (pSpiIo->taskId != AA_INVALID_TASK)
		{
			pSpiIo->result = SPI_EDMA ;
			aaSignalSend (pSpiIo->taskId, AA_SIGNAL_IO) ;
			pSpiIo->taskId = AA_INVALID_TASK ;	// Prevent multiple wake up
		}
	}
	else if (event == DMA_EV_COMPLETE)
	{
		if (pSpiIo->taskId != AA_INVALID_TASK)
		{
			pSpiIo->result = SPI_ENONE ;
dbgPin1 (0u) ;
			aaSignalSend (pSpiIo->taskId, AA_SIGNAL_IO) ;
			pSpiIo->taskId = AA_INVALID_TASK ;	// Prevent multiple wake up
dbgPin1 (1u) ;
		}
	}
	else
	{
	}

	(void) source ;
}

//----------------------------------------------------------------------
//	Configure one DMA stream for SPI

static	void	spiDmaConfigure (const dmaDesc_t * pDmaDesc, spiIo_t * pSpiIo) ;

static	void	spiDmaConfigure (const dmaDesc_t * pDmaDesc, spiIo_t * pSpiIo)
{
	dma_t				* pDma ;
	DMA_Stream_TypeDef	* pStream ;
	SPI_TypeDef 		* pSpi = pSpiIo->pSpi ;
	uint32_t			streamNum ;
	uint32_t			reg ;

	if (pDmaDesc->dmaNum == 0u)
	{
		return ;	// No DMA to configure
	}

	pDma		= DMA_GETDMA (pDmaDesc->dmaNum) ;
	streamNum	= DMA_GETSTREAM (pDmaDesc) ;
	pStream		= dmaStreamAddressRaw (pDma, streamNum) ;

	// Enable DMA clock
	rccEnableClock (pDma, 1u << ((pDma == (dma_t *) DMA1) ? RCC_AHB1ENR_DMA1EN_Pos : RCC_AHB1ENR_DMA2EN_Pos)) ;

	// Reset stream
	dmaStreamResetRaw (pDma, streamNum) ;

	// Configure the DMA/Stream functional parameters
	reg = pStream->CR ;
	reg &= ~(DMA_SxCR_DIR | DMA_SxCR_CIRC | DMA_SxCR_PINC | DMA_SxCR_MINC | DMA_SxCR_PSIZE | DMA_SxCR_MSIZE | DMA_SxCR_PL | DMA_SxCR_PFCTRL) ;
	if ((pSpi->CR1 & SPI_CR1_DFF) == 0u)
	{
		reg |= LL_DMA_PDATAALIGN_BYTE | LL_DMA_MDATAALIGN_BYTE ;
	}
	else
	{
		reg |= LL_DMA_PDATAALIGN_HALFWORD | LL_DMA_MDATAALIGN_HALFWORD ;
	}
	pStream->CR = reg | (pDmaDesc->config         |
						LL_DMA_PERIPH_NOINCREMENT |
						LL_DMA_MEMORY_INCREMENT) ;

	// Set peripheral address
	pStream->PAR = (uint32_t) & pSpi->DR ;

	// Set stream channel
	pStream->CR |= (pStream->CR & ~DMA_SxCR_CHSEL) | (DMA_GETCHANNEL (pDmaDesc) << DMA_SxCR_CHSEL_Pos) ;

	// Configure master SPI DMA interrupts. No DMA interrupt for slave SPI.
	if ((pSpi->CR1 & SPI_CR1_MSTR) != 0)
	{
		dmaSetIrq (pDmaDesc) ;
		dmaSetCallback (pDmaDesc, spiCallback, (uintptr_t) pSpiIo) ;
	}

	// Enable DMA error interrupt
	pStream->CR |= DMA_SxCR_TEIE ;
}

//----------------------------------------------------------------------
//	Initialize SPI device: SPI master or slave mode, polling or DMA, no CRC
//	Speed is a wish, because this SPI only allows a very limited number of speeds
//	Default exchange timeout is SPI_DEF_TIMEOUT ticks

spiStatus_t		spiConfigure_	(const spiDesc_t * pSpiDesc, uint32_t speed, uint32_t flags, spiHandle_t * pHdl)
{
	LL_SPI_InitTypeDef	SPI_InitStruct ;
	spiIo_t				* pSpiIo ;
	SPI_TypeDef			* pSpi = pSpiDesc->pSpi ;

	pSpiIo = spiGetIoDesc (pSpi) ;
	if (pSpiIo == NULL)
	{
		return SPI_EARG ;
	}
	pSpiIo->pSpi		= pSpi ;
	pSpiIo->pSpiDesc	= pSpiDesc ;
	pSpiIo->taskId		= AA_INVALID_TASK ;
	pSpiIo->timeout		= SPI_DEF_TIMEOUT ;

	// Enable the peripheral clock for SPI
	rccEnableClock (pSpi, 1u << pSpiDesc->periphPos) ;

	// Reset peripheral => all configuration registers to reset values
	rccResetPeriph (pSpi, 1u << pSpiDesc->periphPos) ;

	// Configure pins
	gpioConfigurePin (& pSpiDesc->pinMosi) ;
	gpioConfigurePin (& pSpiDesc->pinMiso) ;

	// Set SCK value before pin init: avoid glitch if there is external pull up/down
	if ((flags & SPI_MODE_2) == 0)
	{
		gpioConfigureAndSetPin (& pSpiDesc->pinSck, 0) ;	// SPI_MODE_0 or SPI_MODE_1 : Low level when idle state
	}
	else
	{
		gpioConfigureAndSetPin (& pSpiDesc->pinSck, 1) ;	// SPI_MODE_2 or SPI_MODE_3 : High level when idle state
	}

	// Set NSS value before pin init: avoid glitch if there is external pull up
	gpioConfigureAndSetPin (& pSpiDesc->pinNss, 1) ;

	SPI_InitStruct.Mode				= ((flags & SPI_MASTER_) != 0) ? LL_SPI_MODE_MASTER : LL_SPI_MODE_SLAVE ;
	SPI_InitStruct.NSS				= ((flags & SPI_NSS_MASK) == SPI_NSS_USER) ? LL_SPI_NSS_SOFT : LL_SPI_NSS_HARD_OUTPUT ;
	SPI_InitStruct.BaudRate			= spiSpeedToBr (pSpiDesc, speed) << SPI_CR1_BR_Pos ;
	SPI_InitStruct.CRCCalculation	= LL_SPI_CRCCALCULATION_DISABLE ;
	SPI_InitStruct.CRCPoly			= 10u ;

	// Set SPI polarity / phase
	switch (flags & SPI_MODE_MASK)
	{
		default:
		case SPI_MODE_0:		// Default
			SPI_InitStruct.ClockPolarity	= LL_SPI_POLARITY_LOW ;
			SPI_InitStruct.ClockPhase		= LL_SPI_PHASE_1EDGE ;
			break ;

		case SPI_MODE_1:
			SPI_InitStruct.ClockPolarity	= LL_SPI_POLARITY_LOW ;
			SPI_InitStruct.ClockPhase		= LL_SPI_PHASE_2EDGE ;
			break ;

		case SPI_MODE_2:
			SPI_InitStruct.ClockPolarity	= LL_SPI_POLARITY_HIGH ;
			SPI_InitStruct.ClockPhase		= LL_SPI_PHASE_1EDGE ;
			break ;

		case SPI_MODE_3:
			SPI_InitStruct.ClockPolarity	= LL_SPI_POLARITY_HIGH ;
			SPI_InitStruct.ClockPhase		= LL_SPI_PHASE_2EDGE ;
			break ;
	}

	// Set transfer mode
	switch (flags & SPI_DIR_MASK)
	{
		default:
		case SPI_DIR_FULL_DUPLEX:		// Default
			SPI_InitStruct.TransferDirection	= LL_SPI_FULL_DUPLEX ;
			break ;

		case SPI_DIR_RX_ONLY:
			SPI_InitStruct.TransferDirection	= LL_SPI_SIMPLEX_RX ;
			break ;

		case SPI_DIR_HALF_RX:
			SPI_InitStruct.TransferDirection	= LL_SPI_HALF_DUPLEX_RX ;
			break ;

		case SPI_DIR_HALF_TX:
			SPI_InitStruct.TransferDirection	= LL_SPI_HALF_DUPLEX_TX ;
			break ;
	}

	// Bit order default: MSB
	SPI_InitStruct.BitOrder = ((flags & SPI_LSB_FIRST) != 0u) ? LL_SPI_LSB_FIRST : LL_SPI_MSB_FIRST ;

	// Data width default: 8
	SPI_InitStruct.DataWidth = (flags & SPI_BITS_16) != 0u ? LL_SPI_DATAWIDTH_16BIT : LL_SPI_DATAWIDTH_8BIT ;

	LL_SPI_Init (pSpi, & SPI_InitStruct) ;

	LL_SPI_SetStandard (pSpi, LL_SPI_PROTOCOL_MOTOROLA) ;	// Default mode is Motorola

	// Configure DMA
	if ((flags & SPI_DIR_MASK) == SPI_DIR_FULL_DUPLEX  ||  (flags & SPI_DIR_MASK) == SPI_DIR_HALF_TX)
	{
		// Initialize TX DMA
		spiDmaConfigure (& pSpiIo->pSpiDesc->dmaTx, pSpiIo) ;
	}
	if ((flags & SPI_DIR_MASK) != SPI_DIR_HALF_TX)
	{
		// Initialize RX DMA
		spiDmaConfigure (& pSpiIo->pSpiDesc->dmaRx, pSpiIo) ;
	}

	* pHdl = (spiHandle_t) pSpiIo ;
	return SPI_ENONE ;
}

//----------------------------------------------------------------------
//	Sometime things goes wrong, the SPI TX FIFO is not empty, and the next transmission will be wrong...
//	There is no mean to empty the TX FIFO!
//	So we have to reconfigure the whole SPI which takes time.
//	This function allows you quickly to reset the SPI and reconfigure it identically

void		spiConfigAgain	(const spiHandle_t hdl)
{
	spiIo_t			* pSpiIo = (spiIo_t *) hdl ;
	SPI_TypeDef		* pSpi = pSpiIo->pSpiDesc->pSpi ;
	uint32_t		cr1, cr2 ;

	// Save the configuration registers
	cr1 = pSpi->CR1 ;
	cr2 = pSpi->CR2 ;

	// Reset SPI
	rccResetPeriph (pSpi, 1u << pSpiIo->pSpiDesc->periphPos) ;

	// Restore the configuration
	pSpi->CR1 = cr1 ;
	pSpi->CR2 = cr2 ;
}

//----------------------------------------------------------------------
//	Data exchange with polling
//	Support only full duplex, or half duplex TX
//	pDataOut or pDataIn can be NULL if they are not necessary
//	Count is an item count (not a byte count)

spiStatus_t spiMasterExchange (spiHandle_t hdl, void * pDataOut, void * pDataIn, uint32_t count)
{
	uint32_t		ii ;
	spiIo_t			* pSpiIo = (spiIo_t *) hdl ;
	SPI_TypeDef		* pSpi = pSpiIo->pSpi ;
	
dbgPin1 (1) ;
	// Enable SPI
	pSpi->CR1 |= SPI_CR1_SPE ;

dbgPin1 (0) ;
	// Send/receive data
	if ((pSpi->CR1 & SPI_CR1_DFF) == 0u)
	{
		// 8 bits data
		uint8_t		data;
		uint8_t		* pOut = (uint8_t *) pDataOut ;
		uint8_t		* pIn  = (uint8_t *) pDataIn ;

		if ((pSpi->CR1 & (SPI_CR1_BIDIMODE | SPI_CR1_BIDIOE)) == 0u)
		{
			// RX + TX
			for (ii = 0 ; ii < count ; ii++)
			{
				// Write data to send
				data = 0 ;
				if (pOut != NULL)
				{
					data = * pOut++ ;
				}
				pSpi->DR = data ;

				while ((pSpi->SR & SPI_SR_TXE) == 0)
				{
				}

				// Get received data
				while ((pSpi->SR & SPI_SR_RXNE) == 0)
				{
				}

				data = (uint8_t) pSpi->DR ;
				if (pIn != NULL)
				{
					* pIn++ = data ;
				}
			}
		}
		else
		{
			// TX only
			for (ii = 0 ; ii < count ; ii++)
			{
				// Write data to send
				pSpi->DR = * pOut++ ;

				while ((pSpi->SR & SPI_SR_TXE) == 0)
				{
				}
			}
		}

	}
	else
	{
		// 16 bits data
		uint16_t	data ;
		uint16_t	* pOut = (uint16_t *) pDataOut ;
		uint16_t	* pIn  = (uint16_t *) pDataIn ;

		if ((pSpi->CR1 & (SPI_CR1_BIDIMODE | SPI_CR1_BIDIOE)) == 0u)
		{
			// RX + TX
			for (ii = 0 ; ii < count ; ii++)
			{
				// Write data to send
				data = 0 ;
				if (pOut != NULL)
				{
					data = * pOut++ ;
				}
				pSpi->DR = data ;

				while ((pSpi->SR & SPI_SR_TXE) == 0)
				{
				}

				// Get received data
				while ((pSpi->SR & SPI_SR_RXNE) == 0)
				{
				}

				data = (uint16_t) pSpi->DR ;
				if (pIn != NULL)
				{
					* pIn++ = data ;
				}
			}
		}
		else
		{
			// TX only
			for (ii = 0 ; ii < count ; ii++)
			{
				// Write data to send
				pSpi->DR = * pOut++ ;

				while ((pSpi->SR & SPI_SR_TXE) == 0)
				{
				}
			}
		}
	}
dbgPin1 (1) ;

	// Wait for end of transmission
	while ((pSpi->SR & SPI_SR_BSY) != 0u)
	{
	}

	// Disable SPI
	pSpi->CR1 &= ~SPI_CR1_SPE ;
dbgPin1 (0u) ;

	return SPI_ENONE ;
}

//----------------------------------------------------------------------
//	pDataOut is a pointer to Tx data, can't be NULL
//	pDataIn  is a pointer to Rx data:
//		Full duplex: can be NULL if there is nothing to receive
//		TX only: must be NULL
//	Count is an item count (not a byte count)

spiStatus_t		spiMasterExchangeDma (spiHandle_t hdl, void * pDataOut, void * pDataIn, uint32_t count)
{
	spiIo_t				* pSpiIo = (spiIo_t *) hdl ;
	SPI_TypeDef			* pSpi = pSpiIo->pSpi ;
	dma_t				* pDma ;
	DMA_Stream_TypeDef	* pStreamTx ;
	DMA_Stream_TypeDef	* pStreamRx = NULL ;	// Avoid GCC false message "may be used uninitialized in this function"
	spiStatus_t			sts ;
	aaError_t			res ;

dbgPin1 (1) ;
	// Check if SPI is configured (MSTR bit configured)
	if (pSpi->CR1 == 0u)
	{
		AA_ASSERT (0) ;
		return SPI_ENOTINIT ;
	}

	// pDataOut can't be NULL
	if (pDataOut == NULL)
	{
		AA_ASSERT (0) ;
		return SPI_EARG ;
	}

	// If TX only mode, then pDataIn must be NULL
	if (LL_SPI_GetTransferDirection (pSpi) != LL_SPI_FULL_DUPLEX)
	{
		if (pDataIn != NULL)
		{
			AA_ASSERT (0) ;
			return SPI_EARG ;
		}
	}

	pSpiIo->taskId = aaTaskSelfId () ;
	aaSignalClear (AA_SELFTASKID, AA_SIGNAL_IO) ;

	// Configure and start RX DMA
	if (pDataIn != NULL)
	{
		pDma = DMA_GETDMA (pSpiIo->pSpiDesc->dmaRx.dmaNum) ;
		pStreamRx = & pDma->stream [DMA_GETSTREAM (& pSpiIo->pSpiDesc->dmaRx)]  ;
		pStreamRx->CR		&= ~DMA_SxCR_EN ;
		dmaClearFlagsRaw	(pDma, DMA_GETSTREAM (& pSpiIo->pSpiDesc->dmaRx), DMA_FLAG_ALLIF) ;
		pStreamRx->NDTR		= count ;
		pStreamRx->M0AR		= (uint32_t) pDataIn ;
		pStreamRx->CR		|= DMA_SxCR_TCIE ;	// Enable RX DMA IT transfer complete
		pStreamRx->CR		|= DMA_SxCR_EN ;

		LL_SPI_EnableDMAReq_RX (pSpi) ;
	}

	// Configure and start TX DMA
	pDma = DMA_GETDMA (pSpiIo->pSpiDesc->dmaTx.dmaNum) ;
	pStreamTx = & pDma->stream [DMA_GETSTREAM (& pSpiIo->pSpiDesc->dmaTx)]  ;
	pStreamTx->CR		&= ~DMA_SxCR_EN ;
	dmaClearFlagsRaw	(pDma, DMA_GETSTREAM (& pSpiIo->pSpiDesc->dmaTx), DMA_FLAG_ALLIF) ;
	pStreamTx->NDTR		= count ;
	pStreamTx->M0AR		= (uint32_t) pDataOut ;
	if (pDataIn == NULL)
	{
		// Tx only, enable TX DMA IT transfer complete
		pStreamTx->CR |= DMA_SxCR_TCIE ;
	}
	pStreamTx->CR |= DMA_SxCR_EN ;

	LL_SPI_EnableDMAReq_TX (pSpi) ;

	// Enable SPI
	pSpi->CR1 |= SPI_CR1_SPE ;

	// Wait end of DMA
dbgPin1 (0u) ;
	res = aaSignalWait (AA_SIGNAL_IO, NULL, AA_SIGNAL_AND, pSpiIo->timeout) ;
dbgPin1 (0u) ;
	if (res == AA_ENONE)
	{
		sts = SPI_ENONE ;
	}
	else if (res == AA_ETIMEOUT)
	{
		sts = SPI_ETIMEOUT ;
	}
	else
	{
		sts = SPI_EFAIL ;	// Other error
	}
	if (pSpiIo->result == SPI_EDMA)
	{
		sts = SPI_EDMA ;
	}
	pSpiIo->taskId = AA_INVALID_TASK ;

	// Wait for end of transmission
	if (sts == SPI_ENONE)
	{
		while ((pSpi->SR & SPI_SR_BSY) != 0u)
		{
		}
	}
dbgPin1 (1) ;

	// Disable DMA
	pStreamTx->CR &= ~DMA_SxCR_EN ;
	if (pDataIn != NULL)
	{
		pStreamRx->CR &= ~DMA_SxCR_EN ;
	}

	// Disable SPI
	pSpi->CR1 &= ~SPI_CR1_SPE ;

	LL_SPI_DisableDMAReq_TX (pSpi) ;
	LL_SPI_DisableDMAReq_RX (pSpi) ;

	// Clear SPI SR flags
	{
		volatile uint32_t	reg ;
		reg = pSpi->DR ;
		reg = pSpi->SR ;
		(void) reg ;
	}

dbgPin1 (0u) ;
	return sts ;
}

//----------------------------------------------------------------------
//	pDataOut is a pointer to TX data, can be NULL, ignored if RX only.
//	pDataIn  is a pointer to RX data:
//		Full duplex: can be NULL if there is nothing to receive
//		TX only: must be NULL
//	Count is an item count (not a byte count)

//	This is an asynchronous function: returns after configuring the transaction
//	No DMA interrupts: As the data length is unknown, the DMA buffer is larger than the message
//	The slave inspects the messages on NSS rising (must set an interrupt on NSS rising)

spiStatus_t		spiSlaveExchangeDma (spiHandle_t hdl, void * pDataOut, void * pDataIn, uint32_t count)
{
	spiIo_t				* pSpiIo = (spiIo_t *) hdl ;
	dma_t				* pDma ;
	DMA_Stream_TypeDef		* pStream ;
	uint32_t			txEn = 0 ;
	uint32_t			rxEn = 0 ;
	uint32_t			cr1 ;

dbgPin1 (1) ;
	// Check if SPI is configured as slave
	cr1 = pSpiIo->pSpi->CR1 ;
	if ((cr1 & SPI_CR1_MSTR) != 0)
	{
		AA_ASSERT (0) ;
		return SPI_ENOTINIT ;
	}

	// What direction must be configured ?
	if ((cr1 & SPI_CR1_RXONLY) != 0)
	{
		rxEn = 1 ;		// RX only
	}
	else if ((cr1 & (SPI_CR1_BIDIMODE | SPI_CR1_BIDIOE)) == (SPI_CR1_BIDIMODE | SPI_CR1_BIDIOE))
	{
		txEn = 1 ;		// TX only
	}
	else if ((cr1 & (SPI_CR1_BIDIMODE | SPI_CR1_BIDIOE)) == SPI_CR1_BIDIMODE)
	{
		rxEn = 1 ;		// RX only
	}
	else
	{
		// Full duplex
		if (pDataIn != NULL)
		{
			rxEn = 1 ;
		}
		if (pDataOut != NULL)
		{
			txEn = 1 ;
		}
	}

	// Check data buffer consistency
	if (rxEn == 1  &&  pDataIn == NULL)
	{
		AA_ASSERT (0) ;
		return SPI_EARG ;
	}
	if (txEn == 1  &&  pDataOut == NULL)
	{
		AA_ASSERT (0) ;
		return SPI_EARG ;
	}

	// Configure and start RX DMA
	// What direction must be configured ?
	if ((cr1 & SPI_CR1_RXONLY) != 0)
	{
		rxEn = 1 ;		// RX only
	}
	else if ((cr1 & (SPI_CR1_BIDIMODE | SPI_CR1_BIDIOE)) == (SPI_CR1_BIDIMODE | SPI_CR1_BIDIOE))
	{
		txEn = 1 ;		// TX only
	}
	else if ((cr1 & (SPI_CR1_BIDIMODE | SPI_CR1_BIDIOE)) == SPI_CR1_BIDIMODE)
	{
		rxEn = 1 ;		// RX only
	}
	else
	{
		// Full duplex
		if (pDataIn != NULL)
		{
			rxEn = 1 ;
		}
		if (pDataOut != NULL)
		{
			txEn = 1 ;
		}
	}

	// Check data buffer consistency
	if (rxEn == 1  &&  pDataIn == NULL)
	{
		AA_ASSERT (0) ;
		return SPI_EARG ;
	}
	if (txEn == 1  &&  pDataOut == NULL)
	{
		AA_ASSERT (0) ;
		return SPI_EARG ;
	}

	// Configure and start RX DMA
	if (rxEn == 1)
	{
		AA_ASSERT (pSpiIo->pSpiDesc->dmaRx.dmaNum != DMA_NONE) ;
		if (pSpiIo->pSpiDesc->dmaRx.dmaNum == DMA_NONE)
		{
			return SPI_EARG ;
		}
		pDma = DMA_GETDMA (pSpiIo->pSpiDesc->dmaRx.dmaNum) ;
		pStream = & pDma->stream [DMA_GETSTREAM (& pSpiIo->pSpiDesc->dmaRx)]  ;
		pStream->CR			&= ~DMA_SxCR_EN ;
		dmaClearFlagsRaw	(pDma, DMA_GETSTREAM (& pSpiIo->pSpiDesc->dmaRx), DMA_FLAG_ALLIF) ;
		pStream->NDTR		= count ;
		pStream->M0AR		= (uint32_t) pDataIn ;
		pStream->CR			|= DMA_SxCR_EN ;

		LL_SPI_EnableDMAReq_RX (pSpiIo->pSpi) ;
	}

	// Configure and start TX DMA
	if (txEn == 1)
	{
		AA_ASSERT (pSpiIo->pSpiDesc->dmaTx.dmaNum != DMA_NONE) ;
		if (pSpiIo->pSpiDesc->dmaTx.dmaNum == DMA_NONE)
		{
			return SPI_EARG ;
		}
		pDma = DMA_GETDMA (pSpiIo->pSpiDesc->dmaTx.dmaNum) ;
		pStream = & pDma->stream [DMA_GETSTREAM (& pSpiIo->pSpiDesc->dmaTx)]  ;
		pStream->CR			&= ~DMA_SxCR_EN ;
		dmaClearFlagsRaw	(pDma, DMA_GETSTREAM (& pSpiIo->pSpiDesc->dmaTx), DMA_FLAG_ALLIF) ;
		pStream->NDTR		= count ;
		pStream->M0AR		= (uint32_t) pDataOut ;
		pStream->CR			|= DMA_SxCR_EN ;

		LL_SPI_EnableDMAReq_TX (pSpiIo->pSpi) ;
	}

	// Enable SPI
	pSpiIo->pSpi->CR1 |= SPI_CR1_SPE ;

	return SPI_ENONE ;
}

//----------------------------------------------------------------------
//	To be used on slave side, at the end of an exchange (on the NSS rising edge)
//	Disable DMAs and SPI

spiStatus_t			spiSlaveAbort			(spiHandle_t hdl)
{
	SPI_TypeDef			* pSpi ;
	const spiDesc_t		* pSpiDesc = ((spiIo_t *) hdl)->pSpiDesc ;
	dma_t				* pDma ;
	DMA_Stream_TypeDef	* pStream ;

	// Disable DMA streams
	if (pSpiDesc->dmaRx.dmaNum != DMA_NONE)
	{
		pDma = DMA_GETDMA (pSpiDesc->dmaRx.dmaNum) ;
		pStream = & pDma->stream [DMA_GETSTREAM (& pSpiDesc->dmaRx)]  ;
		pStream->CR	&= ~DMA_SxCR_EN ;
	}
	if (pSpiDesc->dmaTx.dmaNum != DMA_NONE)
	{
		pDma = DMA_GETDMA (pSpiDesc->dmaTx.dmaNum) ;
		pStream = & pDma->stream [DMA_GETSTREAM (& pSpiDesc->dmaTx)]  ;
		pStream->CR	&= ~DMA_SxCR_EN ;
	}

	// Disable SPI
	pSpi = ((spiIo_t *) hdl)->pSpi ;
	pSpi->CR1 &= ~SPI_CR1_SPE ;

	// Disable DMAs
	pSpi->CR2 &= ~(SPI_CR2_RXDMAEN | SPI_CR2_TXDMAEN) ;

	return SPI_ENONE ;
}

//----------------------------------------------------------------------
