/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	isrpipe.h		Communication channel for Zero Latency Interrupts

	When		Who	What
	15/09/19	ac	Creation

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2023 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#if ! defined AA_ISRPIPE_H_
#define AA_ISRPIPE_H_
//--------------------------------------------------------------------------------

// Configure pipe package
#define		AA_PIPE32_MAX			2			// Count of pipe32 to use
#define		AA_INVALID_PIPE32		NULL		// Invalid pipe32 ID

typedef	void *  aaPipe32Id_t ;

//--------------------------------------------------------------------------------
//	Memory blocks pool for Zero Latency Interrupts
//	The pool use uint8_t (to save memory) to store blocks indexes, so the pool is limited to 256 blocks.
//	No protection or blocking: only 1 taker and one giver.

typedef struct aaBlockPool_s
{
	uint8_t			* pBuffer ;		// The ring buffer. Mandatory first item of the struct.
	uint8_t			* pIndex ;		// Arrays of indexes of blocks
	uint16_t		count ;			// Count of blocks in pBuffer
	uint16_t		size ;			// Size of each block
	uint16_t		mask ;
	uint16_t		giveIndex ;		// Index of next block to give to   pIndex array
	uint16_t		takeIndex ;		// Index of next block to take from pIndex array
	uint8_t			flags ;
	uint8_t			underflow ;		// 1 when the taker can't get block (the giver can always give)
	uint32_t		magic ;			// To check if the pool is really a pool

} aaBlockPool_t ;

//--------------------------------------------------------------------------------
#ifdef __cplusplus
extern "C" {
#endif

// It is mandatory to call this function before call to pipe functions
void		aaPipe32Init			(void) ;

aaError_t	aaPipe32Create			(aaPipe32Id_t * pId, uintptr_t * pBuffer, uint32_t size, uint32_t flags) ;
aaError_t	aaPipe32Delete			(aaPipe32Id_t id) ;
aaError_t	aaPipe32Reset			(aaPipe32Id_t id) ;
aaError_t	aaPipe32Give			(aaPipe32Id_t id, uintptr_t data) ;
aaError_t	aaPipe32Take			(aaPipe32Id_t id, uintptr_t * pData) ;
aaError_t	aaPipe32ToGive			(aaPipe32Id_t id, uint32_t * pCount) ;	// Get the number of data that can be written
aaError_t	aaPipe32ToTake			(aaPipe32Id_t id, uint32_t * pCount) ;	// Get the number of data that can be read

// Flags values of aaPipe32Create()
#define		AA_PIPE_ISR_WRITER		0x0001		// The ISR  is the writer, the task is the reader
#define		AA_PIPE_TASK_WRITER		0x0002		// The task is the writer, the ISR  is the reader

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

aaError_t	aaBlockPoolCreate		(aaBlockPool_t * pPool, uint8_t * pBuffer, uint8_t * pIndex, uint32_t count, uint32_t size) ;
aaError_t	aaBlockPoolDelete		(aaBlockPool_t * pPool) ;
aaError_t	aaBlockPoolReset		(aaBlockPool_t * pPool) ;
aaError_t	aaBlockPoolTake			(aaBlockPool_t * pPool, void ** ppBlock) ;
aaError_t	aaBlockPoolGive			(aaBlockPool_t * pPool, void * pBlock) ;
aaError_t	aaBlockPoolToGive		(aaBlockPool_t * pPool, uint32_t * pCount) ;
aaError_t	aaBlockPoolToTake		(aaBlockPool_t * pPool, uint32_t * pCount) ;

#ifdef __cplusplus
}
#endif
//--------------------------------------------------------------------------------
#endif /* AA_ISRPIPE_H_ */

