/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	testzli.c	How to evaluate and use Zero Latency Interrupts (ZLI)

	When		Who	What
	15/09/19	ac	Creation
	30/04/21	ac	Adapted too new extiBasic (1 callback per exti line)

----------------------------------------------------------------------
*/
#include	"aa.h"
#include	"aakernel.h"	// For aaIntEnter()/aaIntExit() in IRQ handler
#include	"aaprintf.h"

#include	"timbasic.h"
#include	"extibasic.h"

#include	"test.h"
#include	"isrpipe.h"

//--------------------------------------------------------------------------------
//	Common variables

static	uint32_t	testNumber ;
static	uint32_t	zliCounter ;
static	uint32_t	zliDelta ;
static	uint32_t	zliDeltaMax ;
static	uint32_t	zliDeltaMin ;
static	uint32_t	zliTs ;
static	aaTaskId_t	zliTaskId ;
static	uint32_t	clearRequest ;

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
/*
ZLI TEST 1

This test is intended to demonstrate that a ZLI is not disturbed by interrupt inhibition of the real-time kernel.
For this, a timer is configured to generate a very high priority interrupt at the frequency of 1 MHz.

The test is performed on a STM32F429 configured at 168 MHz.
The time between interruptions is therefore theoretically 168 core clock ticks.

The interrupt handler measures the time between 2 interrupts using the time stamp functions.
The minimum and maximum times observed are stored.
These times are displayed every second by the user's task.

Results:
A period around 168 core clock ticks as expected, with min = 152 and max = 188 observed on a long time.
Therefore we observe a jitter of 36/2 clock ticks, that is (1/168) * 18 = 0.107 us.
A so low jitter proves that it does not come from a software interaction.

An oscilloscope is connected to the GPIO defined by BSP_OUTPUT0
This shows that the execution time of the handler is 532 ns, ie 53.2% of the CPU time!
So in theory it is possible to reach an interrupt frequency of 1.87MHz for 100% CPU usage.
*/

#define	ZLI1_FREQ			500000
#define	ZLI1_TIM_PRIORITY	(BSP_MAX_INT_PRIO-1)	// The lower priority never disabled

// The timer descriptor: TIM4, with CH1 output on PB6
// Choose a timer who doesn't share its IRQ with another device (TIMx, DACx...)
// In timerbasic.h set AA_WITH_TIM4 to 0, to disable this IRQ handler in timbasic.c
static	const	timDesc_t	timInterrupt1 =
{
	4u,								// Descriptor identifier
	4u,								// Timer number 1..14
	RCC_APB1RSTR_TIM4RST_Pos,		// RCC_APBxRSTR_TIMnRST_Pos
	RCC_APB1ENR_TIM4EN_Pos,			// RCC_APBxRSTR_TIMnEN_Pos
	TIM_CAP_TYPE_GPT1 | TIM_CAP_CHAN_1_4 | TIM_CAP_UP_DOWN | TIM_CAP_DMA | TIM_CAP_ETR, 	// Capabilities flags
	TIM4,							// Pointer to TIM registers
	TIM4_IRQn,
	{
		// Use only CH1
		{	'B',	6u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },	// Ch1
		{	0u,		7u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },	// Ch2
		{	0u,		8u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },	// Ch3
		{	0u,		9u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },	// Ch4
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE },	// Ch1N
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE },	// Ch2N
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE },	// Ch3N
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },	// ETR
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE },	// BKIN
	},
} ;

//--------------------------------------------------------------------------------

void	zliTest1 (void)
{
	const timDesc_t		* pTimGen = & timInterrupt1 ;		// The timer to generate frequency
	uint32_t			ii ;

	ii = rccGetTimerClockFrequency (pTimGen->pTim) ;

	aaPrintf ("\ntestZli 1 : Timer %u  f %u\n", pTimGen->timNumber, ii) ;

	testNumber		= 1 ;
	zliCounter		= 0 ;
	zliDelta		= 0 ;
	zliDeltaMax		= 0 ;
	zliDeltaMin		= 1000000000 ;
	clearRequest	= 0 ;

	// Configure the timer to generate a frequency on CH1 output
	// and the UPDATE interruption, that call the interrupt handler
	timTimeBaseInit			(pTimGen, 0, ZLI1_FREQ, TIM_INIT_ARR_PRELOADEN) ;
	timConfigureIntr		(pTimGen, ZLI1_TIM_PRIORITY, NULL, 0u, TIM_INTR_UPDATE) ;
	timOutputChannelInit	(pTimGen, 1u, TIM_OCMODE_PWM1) ;
	timChannelEnable		(pTimGen, 1u << 1u) ;

	timStart (pTimGen) ;

	// Display info every second
	while (1)
	{
		aaTaskDelay (bspGetTickRate ()) ;
		aaPrintf ("ZLI %u delta %u  Max %u  Min %u\n", zliCounter, zliDelta, zliDeltaMax, zliDeltaMin) ;

		// If the user press the blue button, the counters are reseted.
		if (bspInput (BSP_BUTTON0) == 1)
		{
			// Clear counters
			clearRequest = 1 ;
		}

		if (aaCheckGetChar () != 0)
		{
			timStop (pTimGen) ;
			(void) aaGetChar () ;
			break ;		// Test cancelled by the user
		}
	}
}

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
/*
ZLI TEST 2

This test is intended to show how a ZLI and a task can interact.
In this test the ZLI:
	Use a pool of buffer to get a buffer
	Give the buffer to a aaPipe32Id_t pipe
	Wakes up a task waiting for a signal with extiSwiTrigger()
This is a very realistic application of ZLI

A ZLI can use very few kernel resources, but it can interact easily with a task.

The application has 2 interrupt drivers: the ZLI and the SWI
The sequence of events is as follows:

The handler of the ZLI triggers an SWI with extiSwiTrigger ()
The handler of the SWI sends a signal to the task with aaSignalSend ().
The task is scheduled when it has the highest priority among active tasks.

This example was tested on a STM32F429 with 168 MHz core clock.
The time between the ZLI of the timer and the first instruction of the task is less than 5 microseconds.
 */

#define		ZLI2_SWI_LINE		6		// The number the ZLI handler use to fire the SWI interrupt
#define		ZLI2_EXTI_HANDLER	extiCallback6
#define		ZLI2_SIGNAL			1		// The SWI interrupt handler send this signal to wake up the user task

#define		ZLI2_FREQ			10		// ZLI frequency
#define		ZLI2_TIM_PRIORITY	(BSP_MAX_INT_PRIO-1)	// The lower priority never disabled

// The timer descriptor: TIM4, with output CH1 on PB6
// Choose a timer who doesn't share its IRQ with another device (TIMx, DACx...)
// In timerbasic.h set AA_WITH_TIM4 to 0, to disable this IRQ handler in timbasic.c
static	const	timDesc_t	timInterrupt2 =
{
	4,								// Descriptor identifier
	4,								// Timer number 1..14
	RCC_APB1RSTR_TIM4RST_Pos,		// RCC_APBxRSTR_TIMnRST_Pos
	RCC_APB1ENR_TIM4EN_Pos,			// RCC_APBxRSTR_TIMnEN_Pos
	TIM_CAP_TYPE_GPT1 | TIM_CAP_CHAN_1_4 | TIM_CAP_UP_DOWN | TIM_CAP_DMA | TIM_CAP_ETR, 	// Capabilities flags
	TIM4,							// Pointer to TIM registers
	TIM4_IRQn,
	{
		{	'B',	6u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },	// Ch1
		{	0u,		7u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },	// Ch2
		{	0u,		8u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },	// Ch3
		{	0u,		9u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },	// Ch4
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE },	// Ch1N
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE },	// Ch2N
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE },	// Ch3N
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },	// ETR
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE },	// BKIN
	},
} ;

// - - - - - - - - - - - - - Define the block pool and the pipe with static buffers
#define		ZLI_POOL_COUNT	8		// Count of buffer in the pool, and count of entry in the pipe

static	aaBlockPool_t	zliPool ;			// The pool of buffers

// The data item to use with pool and pipe
typedef	struct
{
	uint32_t	counter ;
	uint32_t	delta ;
	uint32_t	deltaMax ;
	uint32_t	deltaMin ;

} zliItem_t;

static	uint8_t			zliPoolBuffer [ZLI_POOL_COUNT * sizeof (zliItem_t)] ;
static	uint8_t			zliPoolIndex  [ZLI_POOL_COUNT] ;

static	aaPipe32Id_t	zliPipeId ;		// The communication pipe between ZLI (writer) and task (reader)
static	uintptr_t		zliPipeBuffer	[ZLI_POOL_COUNT] ;

//--------------------------------------------------------------------------------
//	Callback of the software interrupt generated by the ZLI

void ZLI2_EXTI_HANDLER (uint32_t flags)
{
	bspOutput (BSP_OUTPUT0, 1u) ;
	aaIntEnter () ;

	(void) flags ;

	if (zliTaskId != 0)
	{
		aaSignalSend (zliTaskId, ZLI2_SIGNAL) ;	// Wake up the application task
	}
	else
	{
		AA_ASSERT (0) ;
	}

	aaIntExit () ;
	bspOutput (BSP_OUTPUT0, 0u) ;
}

//--------------------------------------------------------------------------------

void	zliTest2 (void)
{
	const timDesc_t		* pTimGen = & timInterrupt2 ;		// The timer to generate frequency
	uint32_t			ii ;
	aaError_t			res ;
	zliItem_t			* pItem ;

	ii = rccGetTimerClockFrequency (pTimGen->pTim) ;

	aaPrintf ("\ntestZli No 2 : Timer %u  f %u\n", pTimGen->timNumber, ii) ;

	testNumber	= 2u ;
	zliCounter	= 0u ;
	zliDelta	= 0u ;
	zliDeltaMax	= 0u ;
	zliDeltaMin	= 1000000000u ;

	// Configure the timer to generate a frequency on CH1 output
	// and the UPDATE interruption, that call the callback
	timTimeBaseInit			(pTimGen, 0u, ZLI2_FREQ, TIM_INIT_ARR_PRELOADEN) ;
	timConfigureIntr		(pTimGen, ZLI2_TIM_PRIORITY, NULL, 0u, TIM_INTR_UPDATE) ;
	timOutputChannelInit	(pTimGen, 1u, TIM_OCMODE_PWM1) ;
	timChannelEnable		(pTimGen, 1u << 1u) ;

	// Prepare SWI interrupt to send the signal, and the task to receive the signal
	extiInitSwi (ZLI2_SWI_LINE, BSP_IRQPRIOMIN_PLUS(2)) ;

	zliTaskId = aaTaskSelfId () ;
	aaSignalClear (zliTaskId, ZLI2_SIGNAL) ;

	// Create the pipe and the pool of messages
	aaPipe32Init () ;
	res = aaPipe32Create (& zliPipeId, zliPipeBuffer, ZLI_POOL_COUNT, AA_PIPE_ISR_WRITER) ;
	AA_ASSERT (res == AA_ENONE) ;

	res = aaBlockPoolCreate (& zliPool, zliPoolBuffer, zliPoolIndex, ZLI_POOL_COUNT, sizeof (zliItem_t)) ;
	AA_ASSERT (res == AA_ENONE) ;


	timStart (pTimGen) ;
	while (1)
	{
		// Wait for signal from SWI
		aaSignalWait (ZLI2_SIGNAL, NULL, AA_SIGNAL_AND, AA_INFINITE) ;
		bspOutput (BSP_OUTPUT0, 1u) ;

		// Get buffers from the pipe
		while (aaPipe32Take (zliPipeId, (uintptr_t *) & pItem) == AA_ENONE)
		{
			// Print data
			aaPrintf ("ZLI %u delta %u  Max %u  Min %u\n", pItem->counter, pItem->delta, pItem->deltaMin, pItem->deltaMax) ;

			// Return the buffer to the pool
			res = aaBlockPoolGive (& zliPool, pItem) ;
			AA_ASSERT (res == AA_ENONE) ;
		}

		bspOutput (BSP_OUTPUT0, 0u) ;

		if (aaCheckGetChar () != 0u)
		{
			timStop (pTimGen) ;
			aaPipe32Delete (zliPipeId) ;
			aaBlockPoolDelete (& zliPool) ;
			(void) aaGetChar () ;
			break ;		// Test cancelled by the user
		}
	}
}

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
// ZLI_TEST 3
//
//	This code allows to generate periodical ZLI to check:
//	- There is no drift or jitter between ZLI and the PWM (which is system clock)
//	- Generate a drift between ZLI and systick. This allows to check nested interrupts (systick vs ZLI)
//
//	Observe the behavior with an oscilloscope:
//	-	CH1 of TIM4 for PWM output (hardware generated so no drift or jitter)
//	-	Toggle of BSP_OUTPUT1 for ZLI output: constant delay compared to the PWM output

#define		ZLI3_FREQ			100						// ZLI frequency Hz
#define		ZLI3_TIM_PRIORITY	(BSP_MAX_INT_PRIO-1)	// The lower priority never disabled

// The timer descriptor: TIM4, with output CH1 on PB6
// Choose a timer who doesn't share its IRQ with another device (TIMx, DACx...)
// In timerbasic.h set AA_WITH_TIM4 to 0, to disable this IRQ handler in timbasic.c
static	const	timDesc_t	zliTimDesc =
{
	4u,								// Descriptor identifier
	4u,								// Timer number 1..14
	RCC_APB1RSTR_TIM4RST_Pos,		// RCC_APBxRSTR_TIMnRST_Pos
	RCC_APB1ENR_TIM4EN_Pos,			// RCC_APBxRSTR_TIMnEN_Pos
	TIM_CAP_TYPE_GPT1 | TIM_CAP_CHAN_1_4 | TIM_CAP_UP_DOWN | TIM_CAP_DMA | TIM_CAP_ETR, 	// Capabilities flags
	TIM4,							// Pointer to TIM registers
	TIM4_IRQn,
	{
		{	'B',	6u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },	// Ch1
		{	0u,		7u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },	// Ch2
		{	0u,		8u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },	// Ch3
		{	0u,		9u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },	// Ch4
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE },	// Ch1N
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE },	// Ch2N
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE },	// Ch3N
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL | AA_GPIO_PULL_UP },	// ETR
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALTERNATE },	// BKIN
	},
} ;

//--------------------------------------------------------------------------------

void	zliTest3 (void)
{
	aaPrintf ("\ntestZli No 3 : Timer %u\n", zliTimDesc.timNumber) ;

	testNumber = 3u ;

	// The timer is configured with a frequency not multiple of the system tick
	// This allows to generate a drift between the two clocks
	// Timer clock is 42 MHz
	timTimeBaseInit			(& zliTimDesc, 9, 9332, TIM_INIT_ARR_PRELOADEN | TIM_INIT_FORCE_ARR) ;	// 1000,03571556127 Hz

	// The timer is configured with a frequency which is a sub multiple of the system tick
	// This allows to check there is no drift between the two clocks
//	timTimeBaseInit			(& zliTimDesc, 0, ZLI3_FREQ, TIM_INIT_ARR_PRELOADEN) ;

	timConfigureIntr		(& zliTimDesc, ZLI3_TIM_PRIORITY, NULL, 0, TIM_INTR_UPDATE) ;
	timOutputChannelInit	(& zliTimDesc, 1u, TIM_OCMODE_PWM1) ;
	timChannelEnable		(& zliTimDesc, 1u << 1u) ;
	timStart				(& zliTimDesc) ;

	while (1)
	{
		aaTaskDelay (100u) ;
		if (aaCheckGetChar () != 0u)
		{
			timStop (& zliTimDesc) ;
			(void) aaGetChar () ;
			break ;		// Test cancelled by the user
		}
	}
}

//----------------------------------------------------------------------
//--------------------------------------------------------------------------------
//	The TIM4_IRQHandler is common to all tests

void	TIM4_IRQHandler (void)
{
	// ZLI : No aaIntEnter() / aaIntExit()

	bspOutput (BSP_OUTPUT0, 1u) ;

	if (testNumber == 1u)		// --------------------------------------
	{
		timInterrupt1.pTim->SR &= ~TIM_SR_UIF ;	// clear UIF flag

		if (clearRequest == 1u)
		{
			// Reset memo
			zliDeltaMax = 0u ;
			zliDeltaMin = 1000000000u ;
			clearRequest = 0u ;
		}

		if (zliCounter < 100000u)
		{
			zliTs = bspTsGet () ;
		}
		else
		{
			zliDelta = bspRawTsDelta (& zliTs) ;
			if (zliDelta > zliDeltaMax)
			{
				zliDeltaMax = zliDelta ;
			}
			if (zliDelta < zliDeltaMin)
			{
				zliDeltaMin = zliDelta ;
			}
		}
		zliCounter ++ ;
	}
	else if (testNumber == 2u)		// --------------------------------------
	{
		zliItem_t			* pItem ;

		timInterrupt2.pTim->SR &= ~TIM_SR_UIF ;	// Clear timer UIF flag

		if (zliCounter < 10u)
		{
			zliTs = bspTsGet () ;
		}
		else
		{
			zliDelta = bspRawTsDelta (& zliTs) ;
			if (zliDelta > zliDeltaMax)
			{
				zliDeltaMax = zliDelta ;
			}
			if (zliDelta < zliDeltaMin)
			{
				zliDeltaMin = zliDelta ;
			}
		}

		// Increment counter
		zliCounter ++ ;

		// Get a buffer
		aaBlockPoolTake (& zliPool, (void **) & pItem) ;

		// Set buffer data
		pItem->counter	= zliCounter ;
		pItem->delta	= zliDelta ;
		pItem->deltaMin	= zliDeltaMin ;
		pItem->deltaMax	= zliDeltaMax ;

		// Send buffer to pipe
		aaPipe32Give (zliPipeId, (uintptr_t) pItem) ;

		// Triggers a software interrupt
		extiSwiTrigger (ZLI2_SWI_LINE) ;
	}
	else if (testNumber == 3u)		// --------------------------------------
	{
		zliTimDesc.pTim->SR &= ~TIM_SR_UIF ;	// clear UIF flag
//		bspToggleOutput (BSP_OUTPUT1) ;
	}
	else
	{
	}

	bspOutput (BSP_OUTPUT0, 0u) ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------

void	zliTest (void)
{
	uint32_t			cc ;

	while (1)
	{
		aaPrintf ("ZLI test, h for menu help\n") ;
		cc = aaGetChar () ;
		aaPutChar ((char) cc) ;
		aaPutChar ('\n') ;

		if (cc == 'h')
		{
			aaPrintf ("1  High frequency interrupt\n") ;
			aaPrintf ("2  Communication with a task\n") ;
			aaPrintf ("3  Evaluate ZLI delay/jitter and interrupt nesting\n") ;
			aaPrintf ("q  Quit test\n") ;
			continue ;
		}

		if (cc == 'q')
		{
			break ;	// End of test
		}

		switch (cc)
		{
			case '1':
				zliTest1 () ;
				break ;

			case '2':
				zliTest2 () ;
				break ;

			case '3':
				zliTest3 () ;
				break ;

			default:
				break ;
		}
		if (aaCheckGetChar () != 0u)
		{
			(void) aaGetChar () ;
		}
	}
}

//--------------------------------------------------------------------------------
